from ..clients.metaApi.provisioningProfile_client import ProvisioningProfileDto, ProvisioningProfileClient
from httpx import Response


class ProvisioningProfile:
    """Implements a provisioning profile entity"""
    def __init__(self, data: ProvisioningProfileDto, provisioning_profile_client: ProvisioningProfileClient):
        """Inits a provisioning profile entity.

        Args:
            data: Provisioning profile data.
            provisioning_profile_client: Provisioning profile REST API client.
        """
        self._data = data
        self._provisioningProfileClient = provisioning_profile_client

    @property
    def id(self) -> str:
        """Returns profile id.

        Returns:
            Profile id.
        """
        return self._data['_id']

    @property
    def name(self) -> str:
        """Returns profile name.

        Returns:
            Profile name.
        """
        return self._data['name']

    @property
    def version(self) -> int:
        """Returns profile version. Possible values are 4 and 5.

        Returns:
            Profile version.
        """
        return self._data['version']

    @property
    def status(self) -> str:
        """Returns profile status. Possible values are new and active.

        Returns:
            Profile status.
        """
        return self._data['status']

    @property
    def broker_timezone(self) -> str:
        """Returns broker timezone name from Time Zone Database.

        Returns:
            Broker timezone name.
        """
        return self._data['brokerTimezone']

    @property
    def broker_dst_switch_timezone(self) -> str:
        """Returns broker DST timezone name from Time Zone Database.

        Returns:
            Broker DST switch timezone name.
        """
        return self._data['brokerDSTSwitchTimezone']

    async def reload(self):
        """Reloads provisioning profile from API.

        Returns:
            A coroutine resolving when provisioning profile is updated.
        """
        self._data = await self._provisioningProfileClient.get_provisioning_profile(self.id)

    async def remove(self) -> Response:
        """Removes provisioning profile. The current object instance should be discarded after returned promise
        resolves.

        Returns:
            A coroutine resolving when provisioning profile is removed.
        """
        return await self._provisioningProfileClient.delete_provisioning_profile(self.id)

    async def upload_file(self, file_name: str, file: str or memoryview) -> Response:
        """Uploads a file to provisioning profile.

        Args:
            file_name: Name of the file to upload. Allowed values are servers.dat for MT5 profile, broker.srv for
            MT4 profile.
            file: Path to a file to upload or buffer containing file contents.

        Returns:
            A coroutine which resolves when the file is uploaded.
        """
        return await self._provisioningProfileClient.upload_provisioning_profile_file(self.id, file_name, file)

    async def update(self, profile):
        """Updates provisioning profile.

        Args:
            profile: Provisioning profile update.

        Returns:
            A coroutine resolving when provisioning profile is updated.
        """
        await self._provisioningProfileClient.update_provisioning_profile(self.id, profile)
        await self.reload()



from ..httpClient import HttpClient
from .history_client import HistoryClient
from ...metaApi.models import date, format_date
from datetime import datetime
import pytest
import respx
from httpx import Response
copy_factory_api_url = 'https://trading-api-v1.agiliumtrade.agiliumtrade.ai'
http_client = HttpClient()
history_client = HistoryClient(http_client, 'header.payload.sign')


@pytest.fixture(autouse=True)
async def run_around_tests():
    global http_client
    global history_client
    http_client = HttpClient()
    history_client = HistoryClient(http_client, 'header.payload.sign')


class TestHistoryClient:
    @respx.mock
    @pytest.mark.asyncio
    async def test_retrieve_providers_from_api(self):
        """Should retrieve providers from API."""
        expected = [{
            'id': '577f095ab64b4d1710de34f6a28ab3bd',
            'name': 'First Last',
            'strategies': [{
                'id': 'ABCD',
                'name': 'Test strategy'
            }]
        }]

        rsps = respx.get(f'{copy_factory_api_url}/users/current/providers')\
            .mock(return_value=Response(200, json=expected))
        accounts = await history_client.get_providers()
        assert rsps.calls[0].request.url == f'{copy_factory_api_url}/users/current/providers'
        assert rsps.calls[0].request.method == 'GET'
        assert rsps.calls[0].request.headers['auth-token'] == 'header.payload.sign'
        assert accounts == expected

    @pytest.mark.asyncio
    async def test_not_retrieve_providers_with_account_token(self):
        """Should not retrieve providers from API with account token."""
        history_client = HistoryClient(http_client, 'token')
        try:
            await history_client.get_providers()
        except Exception as err:
            assert err.__str__() == 'You can not invoke get_providers method, because you have connected with ' + \
                                    'account access token. Please use API access token from ' + \
                                    'https://app.metaapi.cloud/token page to invoke this method.'

    @respx.mock
    @pytest.mark.asyncio
    async def test_retrieve_subscribers_from_api(self):
        """Should retrieve subscribers from API."""
        expected = [{
          'id': '577f095ab64b4d1710de34f6a28ab3bd',
          'name': 'First Last',
          'strategies': [{
            'id': 'ABCD',
            'name': 'Test strategy'
          }]
        }]
        rsps = respx.get(f'{copy_factory_api_url}/users/current/subscribers') \
            .mock(return_value=Response(200, json=expected))
        accounts = await history_client.get_subscribers()
        assert rsps.calls[0].request.url == f'{copy_factory_api_url}/users/current/subscribers'
        assert rsps.calls[0].request.method == 'GET'
        assert rsps.calls[0].request.headers['auth-token'] == 'header.payload.sign'
        assert accounts == expected

    @pytest.mark.asyncio
    async def test_not_retrieve_subscribers_with_account_token(self):
        """Should not retrieve subscribers from API with account token."""
        history_client = HistoryClient(http_client, 'token')
        try:
            await history_client.get_subscribers()
        except Exception as err:
            assert err.__str__() == 'You can not invoke get_subscribers method, because you have connected with ' + \
                                    'account access token. Please use API access token from ' + \
                                    'https://app.metaapi.cloud/token page to invoke this method.'

    @respx.mock
    @pytest.mark.asyncio
    async def test_retrieve_strategies_subscribed_from_api(self):
        """Should retrieve strategies subscribed to from API."""
        expected = [{
            'id': 'ABCD',
            'name': 'Test strategy'
        }]
        rsps = respx.get(f'{copy_factory_api_url}/users/current/strategies-subscribed') \
            .mock(return_value=Response(200, json=expected))
        accounts = await history_client.get_strategies_subscribed()
        assert rsps.calls[0].request.url == f'{copy_factory_api_url}/users/current/strategies-subscribed'
        assert rsps.calls[0].request.method == 'GET'
        assert rsps.calls[0].request.headers['auth-token'] == 'header.payload.sign'
        assert accounts == expected

    @pytest.mark.asyncio
    async def test_not_retrieve_strategies_subscribed_to_with_account_token(self):
        """Should not retrieve strategies subscribed to from API with account token."""
        history_client = HistoryClient(http_client, 'token')
        try:
            await history_client.get_strategies_subscribed()
        except Exception as err:
            assert err.__str__() == 'You can not invoke get_strategies_subscribed method, because you have ' +\
                   'connected with account access token. Please use API access token from ' + \
                   'https://app.metaapi.cloud/token page to invoke this method.'

    @respx.mock
    @pytest.mark.asyncio
    async def test_retrieve_provided_strategies_from_api(self):
        """Should retrieve strategies subscribed to from API."""
        expected = [{
            'id': 'ABCD',
            'name': 'Test strategy'
        }]
        rsps = respx.get(f'{copy_factory_api_url}/users/current/provided-strategies') \
            .mock(return_value=Response(200, json=expected))
        accounts = await history_client.get_provided_strategies()
        assert rsps.calls[0].request.url == f'{copy_factory_api_url}/users/current/provided-strategies'
        assert rsps.calls[0].request.method == 'GET'
        assert rsps.calls[0].request.headers['auth-token'] == 'header.payload.sign'
        assert accounts == expected

    @pytest.mark.asyncio
    async def test_not_retrieve_provided_strategies_with_account_token(self):
        """Should not retrieve provided strategies from API with account token."""
        history_client = HistoryClient(http_client, 'token')
        try:
            await history_client.get_provided_strategies()
        except Exception as err:
            assert err.__str__() == 'You can not invoke get_provided_strategies method, because you have ' +\
                   'connected with account access token. Please use API access token from ' + \
                   'https://app.metaapi.cloud/token page to invoke this method.'

    @respx.mock
    @pytest.mark.asyncio
    async def test_retrieve_transactions_for_provided_strategies_from_api(self):
        """Should retrieve transactions performed on provided strategies from API."""
        expected = [{
            'id': '64664661:close',
            'type': 'DEAL_TYPE_SELL',
            'time': '2020-08-02T21:01:01.830Z',
            'accountId': '0123456789abcdef0123456789abcdef0123456789abcdef0123456789abcdef',
            'symbol': 'EURJPY',
            'subscriber': {
                'id': 'subscriberId',
                'name': 'Subscriber'
            },
            'demo': False,
            'provider': {
                'id': 'providerId',
                'name': 'Provider'
            },
            'strategy': {
                'id': 'ABCD'
            },
            'improvement': 0,
            'providerCommission': 0,
            'platformCommission': 0,
            'quantity': -0.04,
            'lotPrice': 117566.08744776,
            'tickPrice': 124.526,
            'amount': -4702.643497910401,
            'commission': -0.14,
            'swap': -0.14,
            'profit': 0.49
        }]
        time_from = datetime.now()
        time_till = datetime.now()
        rsps = respx.get(url__startswith=f'{copy_factory_api_url}/users/current/provided-strategies') \
            .mock(return_value=Response(200, json=expected))
        accounts = await history_client.get_provided_strategies_transactions(time_from, time_till, ['ABCD'],
                                                                             ['subscriberId'], 100, 200)
        assert rsps.calls[0].request.url == f'{copy_factory_api_url}/users/current/provided-strategies/' \
                                            f'transactions?from={format_date(time_from).replace(":", "%3A")}&' \
                                            f'till={format_date(time_till).replace(":", "%3A")}&strategyId=ABCD&' \
                                            f'subscriberId=subscriberId&offset=100&limit=200'
        assert rsps.calls[0].request.method == 'GET'
        assert rsps.calls[0].request.headers['auth-token'] == 'header.payload.sign'
        expected[0]['time'] = date(expected[0]['time'])
        assert accounts == expected

    @pytest.mark.asyncio
    async def test_not_retrieve_transactions_for_provided_strategies_with_account_token(self):
        """Should not retrieve transactions on provided strategies from API with account token."""
        history_client = HistoryClient(http_client, 'token')
        try:
            await history_client.get_provided_strategies_transactions(datetime.now(), datetime.now())
        except Exception as err:
            assert err.__str__() == 'You can not invoke get_provided_strategies_transactions method, because ' + \
                   'you have connected with account access token. Please use API access token from ' + \
                   'https://app.metaapi.cloud/token page to invoke this method.'

    @respx.mock
    @pytest.mark.asyncio
    async def test_retrieve_transactions_for_subscribed_strategies_from_api(self):
        """Should retrieve transactions performed on strategies current user is subscribed to from API."""
        expected = [{
            'id': '64664661:close',
            'type': 'DEAL_TYPE_SELL',
            'time': '2020-08-02T21:01:01.830Z',
            'accountId': '0123456789abcdef0123456789abcdef0123456789abcdef0123456789abcdef',
            'symbol': 'EURJPY',
            'subscriber': {
                'id': 'subscriberId',
                'name': 'Subscriber'
            },
            'demo': False,
            'provider': {
                'id': 'providerId',
                'name': 'Provider'
            },
            'strategy': {
                'id': 'ABCD'
            },
            'improvement': 0,
            'providerCommission': 0,
            'platformCommission': 0,
            'quantity': -0.04,
            'lotPrice': 117566.08744776,
            'tickPrice': 124.526,
            'amount': -4702.643497910401,
            'commission': -0.14,
            'swap': -0.14,
            'profit': 0.49
        }]
        time_from = datetime.now()
        time_till = datetime.now()
        rsps = respx.get(f'{copy_factory_api_url}/users/current/strategies-subscribed/transactions') \
            .mock(return_value=Response(200, json=expected))
        accounts = await history_client.get_strategies_subscribed_transactions(time_from, time_till, ['ABCD'],
                                                                               ['subscriberId'], 100, 200)
        assert rsps.calls[0].request.url == f'{copy_factory_api_url}/users/current/strategies-subscribed/' \
                                            f'transactions?from={format_date(time_from).replace(":", "%3A")}&' \
                                            f'till={format_date(time_till).replace(":", "%3A")}&strategyId=ABCD&' \
                                            f'providerId=subscriberId&offset=100&limit=200'
        assert rsps.calls[0].request.method == 'GET'
        assert rsps.calls[0].request.headers['auth-token'] == 'header.payload.sign'
        expected[0]['time'] = date(expected[0]['time'])
        assert accounts == expected

    @pytest.mark.asyncio
    async def test_not_retrieve_transactions_for_subscribed_strategies_with_account_token(self):
        """Should not retrieve transactions on strategies subscribed to from API with account token."""
        history_client = HistoryClient(http_client, 'token')
        try:
            await history_client.get_strategies_subscribed_transactions(datetime.now(), datetime.now())
        except Exception as err:
            assert err.__str__() == 'You can not invoke get_strategies_subscribed_transactions method, ' + \
                   'because you have connected with account access token. Please use API access token from ' + \
                   'https://app.metaapi.cloud/token page to invoke this method.'



from .memoryHistoryStorageModel import MemoryHistoryStorageModel
import json
import os
import asyncio
from .models import format_date, convert_iso_time_to_date
from typing import List
from datetime import datetime
from copy import deepcopy


def stringify(obj: dict or List) -> str:
    """Helper function to convert an object to string and compress.

    Returns:
        Stringified and compressed object.
    """
    return json.dumps(obj).replace('": ', '":').replace('}, {', '},{').replace(', "', ',"')


class HistoryFileManager:
    """History storage file manager which saves and loads history on disk."""

    def __init__(self, account_id: str, application: str, history_storage: MemoryHistoryStorageModel = None):
        """Constructs the history file manager instance."""
        self._accountId = account_id
        self._application = application
        self._historyStorage = history_storage
        self._dealsSize = []
        self._startNewDealIndex = -1
        self._historyOrdersSize = []
        self._startNewOrderIndex = -1
        self.update_disk_storage_job = None
        self._isUpdating = False

    def start_update_job(self):
        """Starts a job to periodically save history on disk"""

        async def update_job():
            while True:
                await asyncio.sleep(60)
                await self.update_disk_storage()

        if not self.update_disk_storage_job:
            self.update_disk_storage_job = asyncio.create_task(update_job())

    def stop_update_job(self):
        """Stops a job to periodically save history on disk."""

        self.update_disk_storage_job.cancel()
        self.update_disk_storage_job = None

    def get_item_size(self, item: dict) -> int:
        """Helper function to calculate object size in bytes in utf-8 encoding.

        Returns:
            Size of object in bytes.
        """
        return len(stringify(item).encode('utf-8'))

    def set_start_new_order_index(self, index: int):
        """Sets the index of the earliest changed historyOrder record.

        Args:
            index: Index of the earliest changed record.
        """
        if self._startNewOrderIndex > index or self._startNewOrderIndex == -1:
            self._startNewOrderIndex = index

    def set_start_new_deal_index(self, index: int):
        """Sets the index of the earliest changed deal record.

        Args:
            index: Index of the earliest changed record.
        """
        if self._startNewDealIndex > index or self._startNewDealIndex == -1:
            self._startNewDealIndex = index

    async def get_history_from_disk(self):
        """Retrieves history from saved file.

        Returns:
            A coroutine resolving with an object with deals and historyOrders.
        """

        history = {
            'deals': [],
            'historyOrders': [],
            'lastDealTimeByInstanceIndex': {},
            'lastHistoryOrderTimeByInstanceIndex': {}
        }
        try:
            if os.path.isfile(f'.metaapi/{self._accountId}-{self._application}-config.bin'):
                config = json.loads(open(f'.metaapi/{self._accountId}-{self._application}-config.bin').read())
                history['lastDealTimeByInstanceIndex'] = config['lastDealTimeByInstanceIndex']
                history['lastHistoryOrderTimeByInstanceIndex'] = config['lastHistoryOrderTimeByInstanceIndex']
        except Exception as err:
            print(f'[{datetime.now().isoformat()}] Failed to read history storage config of '
                  f'account {self._accountId}', err)
            os.remove(f'.metaapi/{self._accountId}-{self._application}-config.bin')

        try:
            if os.path.isfile(f'.metaapi/{self._accountId}-{self._application}-deals.bin'):
                deals = json.loads(open(f'.metaapi/{self._accountId}-{self._application}-deals.bin').read())
                self._dealsSize = list(map(self.get_item_size, deals))
                for deal in deals:
                    convert_iso_time_to_date(deal)
                history['deals'] = deals
        except Exception as err:
            print(f'[{datetime.now().isoformat()}] Failed to read deals history storage of '
                  f'account {self._accountId}', err)
            os.remove(f'.metaapi/{self._accountId}-{self._application}-deals.bin')

        try:
            if os.path.isfile(f'.metaapi/{self._accountId}-{self._application}-historyOrders.bin'):
                history_orders = json.loads(open(f'.metaapi/{self._accountId}-{self._application}-historyOrders.bin')
                                            .read())
                self._historyOrdersSize = list(map(self.get_item_size, history_orders))
                for history_order in history_orders:
                    convert_iso_time_to_date(history_order)
                history['historyOrders'] = history_orders
        except Exception as err:
            print(f'[{datetime.now().isoformat()}] Failed to read historyOrders history storage of '
                  f'account {self._accountId}', err)
            os.remove(f'.metaapi/{self._accountId}-{self._application}-historyOrders.bin')
        return history

    async def update_disk_storage(self):
        """Saves unsaved history items to disk storage.

        Returns:
            A coroutine resolving when the history is saved to disk.
        """
        account_id = self._accountId
        application = self._application
        if not os.path.exists('.metaapi'):
            os.mkdir('.metaapi')

        async def replace_records(history_type, start_index: int, replace_items: List, size_array: List) -> List[int]:
            replace_items = self._prepare_save_data(replace_items)
            file_path = f'.metaapi/{account_id}-{application}-{history_type}.bin'
            file_size = os.path.getsize(file_path)
            if start_index == 0:
                f = open(file_path, 'w+')
                f.write(stringify(replace_items))
                f.close()
            else:
                f = open(file_path, 'a+')
                replaced_items = size_array[start_index:]
                start_position = file_size - len(replaced_items) - sum(replaced_items) - 1
                f.seek(start_position)
                f.truncate()
                f.close()
                f = open(file_path, "a+")
                f.write(',' + stringify(replace_items)[1:])
                f.close()
            return size_array[0:start_index] + list(map(self.get_item_size, replace_items))

        if not self._isUpdating:
            self._isUpdating = True
            try:
                await self._update_config()
                if self._startNewDealIndex != -1:
                    if not os.path.isfile(f'.metaapi/{account_id}-{application}-deals.bin'):
                        try:
                            f = open(f'.metaapi/{account_id}-{application}-deals.bin', "w+")
                            f.write(stringify(self._prepare_save_data(self._historyStorage.deals)))
                            f.close()
                        except Exception as err:
                            print(f'[{datetime.now().isoformat()}] Error saving deals on disk for account '
                                  f'{self._accountId}', err)
                        self._dealsSize = list(map(self.get_item_size,
                                                   self._prepare_save_data(self._historyStorage.deals)))
                    else:
                        replace_deals = self._historyStorage.deals[self._startNewDealIndex:]
                        self._dealsSize = await replace_records('deals', self._startNewDealIndex, replace_deals,
                                                                self._dealsSize)
                    self._startNewDealIndex = -1

                if self._startNewOrderIndex != -1:
                    if not os.path.isfile(f'.metaapi/{account_id}-{application}-historyOrders.bin'):
                        try:
                            f = open(f'.metaapi/{account_id}-{application}-historyOrders.bin', "w+")
                            f.write(stringify(self._prepare_save_data(self._historyStorage.history_orders)))
                            f.close()
                        except Exception as err:
                            print(f'[{datetime.now().isoformat()}] Error saving historyOrders on disk for '
                                  f'account {account_id}', err)
                        self._historyOrdersSize = list(map(
                            self.get_item_size, self._prepare_save_data(self._historyStorage.history_orders)))
                    else:
                        replace_orders = self._historyStorage.history_orders[self._startNewOrderIndex:]
                        self._historyOrdersSize = await replace_records('historyOrders', self._startNewOrderIndex,
                                                                        replace_orders, self._historyOrdersSize)
                    self._startNewOrderIndex = -1
            except Exception as err:
                print(f'[{datetime.now().isoformat()}] Error updating disk storage for '
                      f'account {account_id}', err)
            self._isUpdating = False

    async def _update_config(self):
        """Updates stored config for account."""
        account_id = self._accountId
        history_storage = self._historyStorage
        file_path = f'.metaapi/{account_id}-{self._application}-config.bin'
        try:
            config = {
                'lastDealTimeByInstanceIndex': history_storage.last_deal_time_by_instance_index,
                'lastHistoryOrderTimeByInstanceIndex': history_storage.last_history_order_time_by_instance_index
            }
            f = open(file_path, 'w+')
            f.write(stringify(config))
            f.close()
        except Exception as err:
            print(f'[{datetime.now().isoformat()}] Error updating disk storage config for '
                  f'account {account_id}', err)

    async def delete_storage_from_disk(self):
        """Deletes storage files from disk.

        Returns:
            A coroutine resolving when the history is deleted from disk.
        """
        if os.path.isfile(f'.metaapi/{self._accountId}-{self._application}-config.bin'):
            os.remove(f'.metaapi/{self._accountId}-{self._application}-config.bin')
        if os.path.isfile(f'.metaapi/{self._accountId}-{self._application}-deals.bin'):
            os.remove(f'.metaapi/{self._accountId}-{self._application}-deals.bin')
        if os.path.isfile(f'.metaapi/{self._accountId}-{self._application}-historyOrders.bin'):
            os.remove(f'.metaapi/{self._accountId}-{self._application}-historyOrders.bin')

    def _prepare_save_data(self, arr: List[dict]):
        arr = deepcopy(arr)

        def convert_dates(item):
            for key in item:
                if isinstance(item[key], datetime):
                    item[key] = format_date(item[key])
                elif isinstance(item[key], dict):
                    convert_dates(item[key])

        for item in arr:
            convert_dates(item)
        return arr



from ..httpClient import HttpClient
from .configuration_client import ConfigurationClient
from ...metaApi.models import date
import pytest
import json
from mock import AsyncMock
from copy import deepcopy
import respx
from httpx import Response

copy_factory_api_url = 'https://trading-api-v1.agiliumtrade.agiliumtrade.ai'
http_client = HttpClient()
copy_factory_client = ConfigurationClient(http_client, 'header.payload.sign')


@pytest.fixture(autouse=True)
async def run_around_tests():
    global http_client
    global copy_factory_client
    http_client = HttpClient()
    copy_factory_client = ConfigurationClient(http_client, 'header.payload.sign')


class TestConfigurationClient:
    @pytest.mark.asyncio
    async def test_generate_account_id(self):
        """Should generate account id."""
        assert len(copy_factory_client.generate_account_id()) == 64

    @respx.mock
    @pytest.mark.asyncio
    async def test_update_copyfactory_account(self):
        """Should update CopyFactory account via API."""
        account = {
            'name': 'Demo account',
            'connectionId': 'e8867baa-5ec2-45ae-9930-4d5cea18d0d6',
            'reservedMarginFraction': 0.25,
            'subscriptions': [
                {
                    'strategyId': 'ABCD',
                    'multiplier': 1
                }
            ]
        }
        rsps = respx.put(f'{copy_factory_api_url}/users/current/configuration/accounts/' +
                         '0123456789abcdef0123456789abcdef0123456789abcdef0123456789abcdef',
                         json=account).mock(return_value=Response(200))
        await copy_factory_client.update_account('0123456789abcdef0123456789abcdef0123456789abcdef0123456789abcdef',
                                                 account)
        assert rsps.calls[0].request.url == f'{copy_factory_api_url}/users/current/configuration/accounts/' + \
                                            '0123456789abcdef0123456789abcdef0123456789abcdef0123456789abcdef'
        assert rsps.calls[0].request.method == 'PUT'
        assert rsps.calls[0].request.headers['auth-token'] == 'header.payload.sign'
        assert rsps.calls[0].request.read() == json.dumps(account).encode('utf-8')

    @pytest.mark.asyncio
    async def test_not_update_copyfactory_account_with_account_token(self):
        """Should not update CopyFactory account via API with account token."""
        copy_factory_client = ConfigurationClient(http_client, 'token')
        try:
            await copy_factory_client.update_account('id', {})
        except Exception as err:
            assert err.__str__() == 'You can not invoke update_account method, because you have connected with ' + \
                                    'account access token. Please use API access token from ' + \
                                    'https://app.metaapi.cloud/token page to invoke this method.'

    @respx.mock
    @pytest.mark.asyncio
    async def test_retrieve_copyfactory_accounts_from_api(self):
        """Should retrieve CopyFactory accounts from API."""
        expected = [{
          '_id': '0123456789abcdef0123456789abcdef0123456789abcdef0123456789abcdef',
          'subscriberId': 'subscriberId',
          'name': 'Demo account',
          'connectionId': 'e8867baa-5ec2-45ae-9930-4d5cea18d0d6',
          'reservedMarginFraction': 0.25,
          'subscriptions': [
            {
              'strategyId': 'ABCD',
              'multiplier': 1
            }
          ]
        }]
        rsps = respx.get(f'{copy_factory_api_url}/users/current/configuration/accounts')\
            .mock(return_value=Response(200, json=expected))
        accounts = await copy_factory_client.get_accounts()
        assert rsps.calls[0].request.url == f'{copy_factory_api_url}/users/current/configuration/accounts'
        assert rsps.calls[0].request.method == 'GET'
        assert rsps.calls[0].request.headers['auth-token'] == 'header.payload.sign'
        assert accounts == expected

    @pytest.mark.asyncio
    async def test_not_retrieve_copyfactory_accounts_with_account_token(self):
        """Should not retrieve CopyFactory accounts from API with account token."""
        copy_factory_client = ConfigurationClient(http_client, 'token')
        try:
            await copy_factory_client.get_accounts()
        except Exception as err:
            assert err.__str__() == 'You can not invoke get_accounts method, because you have connected with ' + \
                                    'account access token. Please use API access token from ' + \
                                    'https://app.metaapi.cloud/token page to invoke this method.'

    @respx.mock
    @pytest.mark.asyncio
    async def test_remove_copyfactory_account(self):
        """Should remove CopyFactory account via API."""
        rsps = respx.delete(f'{copy_factory_api_url}/users/current/configuration/accounts/' +
                            '0123456789abcdef0123456789abcdef0123456789abcdef0123456789abcdef')\
            .mock(return_value=Response(204))
        await copy_factory_client\
            .remove_account('0123456789abcdef0123456789abcdef0123456789abcdef0123456789abcdef')
        assert rsps.calls[0].request.url == f'{copy_factory_api_url}/users/current/configuration/accounts/' + \
               '0123456789abcdef0123456789abcdef0123456789abcdef0123456789abcdef'
        assert rsps.calls[0].request.method == 'DELETE'
        assert rsps.calls[0].request.headers['auth-token'] == 'header.payload.sign'

    @pytest.mark.asyncio
    async def test_not_remove_copyfactory_account_with_account_token(self):
        """Should not remove CopyFactory account via API with account token."""
        copy_factory_client = ConfigurationClient(http_client, 'token')
        try:
            await copy_factory_client.remove_account('id')
        except Exception as err:
            assert err.__str__() == 'You can not invoke remove_account method, because you have connected with ' + \
                                    'account access token. Please use API access token from ' + \
                                    'https://app.metaapi.cloud/token page to invoke this method.'

    @respx.mock
    @pytest.mark.asyncio
    async def test_generate_strategy_id(self):
        """Should retrieve CopyFactory accounts from API."""
        expected = {
            'id': 'ABCD'
        }
        rsps = respx.get(f'{copy_factory_api_url}/users/current/configuration/unused-strategy-id')\
            .mock(return_value=Response(200, json=expected))
        id = await copy_factory_client.generate_strategy_id()
        assert rsps.calls[0].request.url == f'{copy_factory_api_url}/users/current/configuration/unused-strategy-id'
        assert rsps.calls[0].request.method == 'GET'
        assert rsps.calls[0].request.headers['auth-token'] == 'header.payload.sign'
        assert id == expected

    @pytest.mark.asyncio
    async def test_not_generate_strategy_id_with_account_token(self):
        """Should not generate strategy id with account token."""
        copy_factory_client = ConfigurationClient(http_client, 'token')
        try:
            await copy_factory_client.generate_strategy_id()
        except Exception as err:
            assert err.__str__() == 'You can not invoke generate_strategy_id method, because you have connected ' + \
                   'with account access token. Please use API access token from ' + \
                   'https://app.metaapi.cloud/token page to invoke this method.'

    @respx.mock
    @pytest.mark.asyncio
    async def test_update_strategy(self):
        """Should update strategy via API."""
        strategy = {
            'name': 'Test strategy',
            'positionLifecycle': 'hedging',
            'connectionId': 'e8867baa-5ec2-45ae-9930-4d5cea18d0d6',
            'maxTradeRisk': 0.1,
            'stopOutRisk': {
                'value': 0.4,
                'startTime': '2020-08-24T00:00:00.000Z'
            },
            'timeSettings': {
                'lifetimeInHours': 192,
                'openingIntervalInMinutes': 5
            }
        }
        rsps = respx.put(f'{copy_factory_api_url}/users/current/configuration/strategies/ABCD', json=strategy)\
            .mock(return_value=Response(200))
        await copy_factory_client.update_strategy('ABCD', strategy)
        assert rsps.calls[0].request.url == f'{copy_factory_api_url}/users/current/configuration/strategies/ABCD'
        assert rsps.calls[0].request.method == 'PUT'
        assert rsps.calls[0].request.headers['auth-token'] == 'header.payload.sign'
        assert rsps.calls[0].request.content == json.dumps(strategy).encode('utf-8')

    @pytest.mark.asyncio
    async def test_not_update_strategy_with_account_token(self):
        """Should not update strategy with account token."""
        copy_factory_client = ConfigurationClient(http_client, 'token')
        try:
            await copy_factory_client.update_strategy('ABCD', {})
        except Exception as err:
            assert err.__str__() == 'You can not invoke update_strategy method, because you have connected with ' + \
                                    'account access token. Please use API access token from ' + \
                                    'https://app.metaapi.cloud/token page to invoke this method.'

    @respx.mock
    @pytest.mark.asyncio
    async def test_retrieve_strategies_from_api(self):
        """Should retrieve strategies from API."""
        expected = [{
          '_id': 'ABCD',
          'providerId': 'providerId',
          'platformCommissionRate': 0.01,
          'name': 'Test strategy',
          'positionLifecycle': 'hedging',
          'connectionId': 'e8867baa-5ec2-45ae-9930-4d5cea18d0d6',
          'maxTradeRisk': 0.1,
          'stopOutRisk': {
            'value': 0.4,
            'startTime': '2020-08-24T00:00:00.000Z'
          },
          'timeSettings': {
            'lifetimeInHours': 192,
            'openingIntervalInMinutes': 5
          }
        }]
        rsps = respx.get(f'{copy_factory_api_url}/users/current/configuration/strategies') \
            .mock(return_value=Response(200, json=expected))
        strategies = await copy_factory_client.get_strategies()
        assert rsps.calls[0].request.url == f'{copy_factory_api_url}/users/current/configuration/strategies'
        assert rsps.calls[0].request.method == 'GET'
        assert rsps.calls[0].request.headers['auth-token'] == 'header.payload.sign'
        assert strategies == expected

    @pytest.mark.asyncio
    async def test_not_retrieve_strategies_with_account_token(self):
        """Should not retrieve strategies with account token."""
        copy_factory_client = ConfigurationClient(http_client, 'token')
        try:
            await copy_factory_client.get_strategies()
        except Exception as err:
            assert err.__str__() == 'You can not invoke get_strategies method, because you have connected with ' + \
                                    'account access token. Please use API access token from ' + \
                                    'https://app.metaapi.cloud/token page to invoke this method.'

    @respx.mock
    @pytest.mark.asyncio
    async def test_remove_strategy(self):
        """Should remove strategy via API."""
        rsps = respx.delete(f'{copy_factory_api_url}/users/current/configuration/strategies/ABCD') \
            .mock(return_value=Response(204))
        await copy_factory_client.remove_strategy('ABCD')
        assert rsps.calls[0].request.url == f'{copy_factory_api_url}/users/current/configuration/strategies/ABCD'
        assert rsps.calls[0].request.method == 'DELETE'
        assert rsps.calls[0].request.headers['auth-token'] == 'header.payload.sign'

    @pytest.mark.asyncio
    async def test_not_remove_strategy_with_account_token(self):
        """Should not remove strategy with account token."""
        copy_factory_client = ConfigurationClient(http_client, 'token')
        try:
            await copy_factory_client.remove_strategy('ABCD')
        except Exception as err:
            assert err.__str__() == 'You can not invoke remove_strategy method, because you have connected with ' + \
                                    'account access token. Please use API access token from ' + \
                                    'https://app.metaapi.cloud/token page to invoke this method.'

    @respx.mock
    @pytest.mark.asyncio
    async def test_retrieve_portfolio_strategies(self):
        """Should retrieve portfolio strategies from API."""
        expected = [{
            '_id': 'ABCD',
            'providerId': 'providerId',
            'platformCommissionRate': 0.01,
            'name': 'Test strategy',
            'members': [{
                'strategyId': 'BCDE'
            }],
            'maxTradeRisk': 0.1,
            'stopOutRisk': {
                'value': 0.4,
                'startTime': '2020-08-24T00:00:00.000Z'
            }
        }]
        rsps = respx.get(f'{copy_factory_api_url}/users/current/configuration/portfolio-strategies') \
            .mock(return_value=Response(200, json=expected))
        strategies = await copy_factory_client.get_portfolio_strategies()
        assert rsps.calls[0].request.url == f'{copy_factory_api_url}/users/current/configuration/' \
                                            f'portfolio-strategies'
        assert rsps.calls[0].request.method == 'GET'
        assert rsps.calls[0].request.headers['auth-token'] == 'header.payload.sign'
        assert strategies == expected

    @pytest.mark.asyncio
    async def test_not_retrieve_portfolio_strategies_with_account_token(self):
        """Should not retrieve portfolio strategies from API with account token."""
        copy_factory_client = ConfigurationClient(http_client, 'token')
        try:
            await copy_factory_client.get_portfolio_strategies()
        except Exception as err:
            assert err.__str__() == 'You can not invoke get_portfolio_strategies method, because you have connected '\
                                    'with account access token. Please use API access token from '\
                                    'https://app.metaapi.cloud/token page to invoke this method.'

    @respx.mock
    @pytest.mark.asyncio
    async def test_update_portfolio_strategy(self):
        """Should update portfolio strategy via API."""
        strategy = {
            'name': 'Test strategy',
            'members': [{
                'strategyId': 'BCDE'
            }],
            'maxTradeRisk': 0.1,
            'stopOutRisk': {
                'value': 0.4,
                'startTime': '2020-08-24T00:00:00.000Z'
            }
        }
        rsps = respx.put(f'{copy_factory_api_url}/users/current/configuration/portfolio-strategies/ABCD',
                         json=strategy).mock(return_value=Response(200))
        await copy_factory_client.update_portfolio_strategy('ABCD', strategy)
        assert rsps.calls[0].request.url == f'{copy_factory_api_url}/users/current/configuration/' \
                                            f'portfolio-strategies/ABCD'
        assert rsps.calls[0].request.method == 'PUT'
        assert rsps.calls[0].request.headers['auth-token'] == 'header.payload.sign'
        assert rsps.calls[0].request.content == json.dumps(strategy).encode('utf-8')

    @pytest.mark.asyncio
    async def test_not_update_portfolio_strategy_with_account_token(self):
        """Should not update portfolio strategy via API with account token."""
        copy_factory_client = ConfigurationClient(http_client, 'token')
        try:
            await copy_factory_client.update_portfolio_strategy('ABCD', {})
        except Exception as err:
            assert err.__str__() == 'You can not invoke update_portfolio_strategy method, because you have connected ' \
                                    'with account access token. Please use API access token from ' \
                                    'https://app.metaapi.cloud/token page to invoke this method.'

    @respx.mock
    @pytest.mark.asyncio
    async def test_remove_portfolio_strategy(self):
        """Should remove portfolio strategy via API."""
        rsps = respx.delete(f'{copy_factory_api_url}/users/current/configuration/portfolio-strategies/ABCD')\
            .mock(return_value=Response(204))
        await copy_factory_client.remove_portfolio_strategy('ABCD')
        assert rsps.calls[0].request.url == f'{copy_factory_api_url}/users/current/configuration/' \
                                            f'portfolio-strategies/ABCD'
        assert rsps.calls[0].request.method == 'DELETE'
        assert rsps.calls[0].request.headers['auth-token'] == 'header.payload.sign'

    @pytest.mark.asyncio
    async def test_not_remove_portfolio_strategy_with_account_token(self):
        """Should not remove portfolio strategy with account token."""
        copy_factory_client = ConfigurationClient(http_client, 'token')
        try:
            await copy_factory_client.remove_portfolio_strategy('ABCD')
        except Exception as err:
            assert err.__str__() == 'You can not invoke remove_portfolio_strategy method, because you have ' \
                                    'connected with account access token. Please use API access token from ' + \
                                    'https://app.metaapi.cloud/token page to invoke this method.'

    @respx.mock
    @pytest.mark.asyncio
    async def test_retrieve_active_resync_tasks_from_api(self):
        """Should retrieve active resynchronization tasks via API."""
        result = [{
            '_id': 'ABCD',
            'type': 'CREATE_STRATEGY',
            'createdAt': '2020-08-25T00:00:00.000Z',
            'status': 'EXECUTING'
        }]
        expected = [{
            '_id': 'ABCD',
            'type': 'CREATE_STRATEGY',
            'createdAt': date('2020-08-25T00:00:00.000Z'),
            'status': 'EXECUTING'
        }]
        rsps = respx.get(f'{copy_factory_api_url}/users/current/configuration/connections/' +
                         'accountId/active-resynchronization-tasks') \
            .mock(return_value=Response(200, json=result))
        strategies = await copy_factory_client.get_active_resynchronization_tasks('accountId')
        assert rsps.calls[0].request.url == f'{copy_factory_api_url}/users/current/configuration/connections/' + \
            'accountId/active-resynchronization-tasks'
        assert rsps.calls[0].request.method == 'GET'
        assert rsps.calls[0].request.headers['auth-token'] == 'header.payload.sign'
        assert strategies == expected

    @pytest.mark.asyncio
    async def test_not_retrieve_resync_tasks_with_account_token(self):
        """Should not retrieve active resynchronization tasks with account token."""
        copy_factory_client = ConfigurationClient(http_client, 'token')
        try:
            await copy_factory_client.get_active_resynchronization_tasks('ABCD')
        except Exception as err:
            assert err.__str__() == 'You can not invoke get_active_resynchronization_tasks method, ' + \
                   'because you have connected with account access token. Please use API access token from ' + \
                   'https://app.metaapi.cloud/token page to invoke this method.'

    @pytest.mark.asyncio
    async def test_wait_resync_tasks(self):
        """Should wait until active resynchronization tasks are completed."""
        active_tasks = [{
            '_id': 'ABCD',
            'type': 'CREATE_STRATEGY',
            'createdAt': '2020-08-25T00:00:00.000Z',
            'status': 'EXECUTING'
        }]
        copy_factory_client._httpClient.request = AsyncMock(side_effect=[deepcopy(active_tasks),
                                                                         deepcopy(active_tasks), []])
        await copy_factory_client.wait_resynchronization_tasks_completed('accountId', 1, 50)
        copy_factory_client._httpClient.request.assert_called_with({
            'url': f"{copy_factory_api_url}/users/current/configuration/connections/accountId" +
                   "/active-resynchronization-tasks",
            'method': 'GET',
            'headers': {
                'auth-token': 'header.payload.sign'
            }
        })
        assert copy_factory_client._httpClient.request.call_count == 3

    @pytest.mark.asyncio
    async def test_timeout_resync_tasks(self):
        """Should time out waiting for active resynchronization tasks to be completed."""
        active_tasks = [{
            '_id': 'ABCD',
            'type': 'CREATE_STRATEGY',
            'createdAt': '2020-08-25T00:00:00.000Z',
            'status': 'EXECUTING'
        }]
        copy_factory_client._httpClient.request = AsyncMock(side_effect=[deepcopy(active_tasks),
                                                                         deepcopy(active_tasks),
                                                                         deepcopy(active_tasks)])
        try:
            await copy_factory_client.wait_resynchronization_tasks_completed('accountId', 0.5, 300)
            raise Exception('TimeoutException is expected')
        except Exception as err:
            assert err.__class__.__name__ == 'TimeoutException'
        copy_factory_client._httpClient.request.assert_called_with({
            'url': f"{copy_factory_api_url}/users/current/configuration/connections/accountId" +
                   "/active-resynchronization-tasks",
            'method': 'GET',
            'headers': {
                'auth-token': 'header.payload.sign'
            }
        })



from .historyStorage import HistoryStorage
from typing import Optional
from abc import ABC, abstractmethod
from datetime import datetime


class MetatraderAccountModel(ABC):
    """Defines interface for a MetaTrader account class."""

    @property
    @abstractmethod
    def id(self) -> str:
        """Returns account id.

        Returns:
            Account id.
        """

    @property
    @abstractmethod
    def name(self) -> str:
        """Returns account name.

        Returns:
            Account name.
        """

    @property
    @abstractmethod
    def type(self) -> str:
        """Returns account type. Possible values are cloud and self-hosted.

        Returns:
            Account type.
        """

    @property
    @abstractmethod
    def login(self) -> str:
        """Returns account login.

        Returns:
            Account login.
        """

    @property
    @abstractmethod
    def server(self) -> str:
        """Returns MetaTrader server which hosts the account.

        Returns:
            MetaTrader server which hosts the account.
        """

    @property
    @abstractmethod
    def provisioning_profile_id(self):
        """Returns id of the account's provisioning profile.

        Returns:
            Id of the account's provisioning profile.
        """

    @property
    @abstractmethod
    def application(self):
        """Returns application name to connect the account to. Currently allowed values are MetaApi and AgiliumTrade.

        Returns:
            Application name to connect the account to.
        """

    @property
    @abstractmethod
    def magic(self):
        """Returns MetaTrader magic to place trades using.

        Returns:
            MetaTrader magic to place trades using.
        """

    @property
    @abstractmethod
    def state(self):
        """Returns account deployment state. One of CREATED, DEPLOYING, DEPLOYED, UNDEPLOYING, UNDEPLOYED, DELETING

        Returns:
            Account deployment state.
        """

    @property
    @abstractmethod
    def connection_status(self):
        """Returns terminal & broker connection status, one of CONNECTED, DISCONNECTED, DISCONNECTED_FROM_BROKER

        Returns:
            Terminal & broker connection status.
        """

    @property
    @abstractmethod
    def access_token(self):
        """Returns authorization access token to be used for accessing single account data.
        Intended to be used in browser API.

        Returns:
            Authorization token.
        """

    @abstractmethod
    async def reload(self):
        """Reloads MetaTrader account from API.

        Returns:
            A coroutine resolving when MetaTrader account is updated.
        """

    @abstractmethod
    async def remove(self):
        """Removes MetaTrader account. Cloud account transitions to DELETING state.
        It takes some time for an account to be eventually deleted. Self-hosted account is deleted immediately.

        Returns:
            A coroutine resolving when account is scheduled for deletion.
        """

    @abstractmethod
    async def deploy(self):
        """Schedules account for deployment. It takes some time for API server to be started and account to reach the
        DEPLOYED state.

        Returns:
            A coroutine resolving when account is scheduled for deployment.
        """

    @abstractmethod
    async def undeploy(self):
        """Schedules account for undeployment. It takes some time for API server to be stopped and account to reach the
        UNDEPLOYED state.

        Returns:
            A coroutine resolving when account is scheduled for undeployment.
        """

    @abstractmethod
    async def redeploy(self):
        """Schedules account for redeployment. It takes some time for API server to be restarted and account to reach
        the DEPLOYED state.

        Returns:
            A coroutine resolving when account is scheduled for redeployment.
        """

    @abstractmethod
    async def wait_deployed(self, timeout_in_seconds=300, interval_in_milliseconds=1000):
        """Waits until API server has finished deployment and account reached the DEPLOYED state.

        Args:
            timeout_in_seconds: Wait timeout in seconds, default is 5m.
            interval_in_milliseconds: Interval between account reloads while waiting for a change, default is 1s.

        Returns:
            A coroutine which resolves when account is deployed.

        Raises:
            TimeoutException: If account has not reached the DEPLOYED state within timeout allowed.
        """

    @abstractmethod
    async def wait_undeployed(self, timeout_in_seconds=300, interval_in_milliseconds=1000):
        """Waits until API server has finished undeployment and account reached the UNDEPLOYED state.

        Args:
            timeout_in_seconds: Wait timeout in seconds, default is 5m.
            interval_in_milliseconds: Interval between account reloads while waiting for a change, default is 1s.

        Returns:
            A coroutine which resolves when account is undeployed.

        Raises:
            TimeoutException: If account have not reached the UNDEPLOYED state within timeout allowed.
        """

    @abstractmethod
    async def wait_removed(self, timeout_in_seconds=300, interval_in_milliseconds=1000):
        """Waits until account has been deleted.

        Args:
            timeout_in_seconds: Wait timeout in seconds, default is 5m.
            interval_in_milliseconds: Interval between account reloads while waiting for a change, default is 1s.

        Returns:
            A coroutine which resolves when account is deleted.

        Raises:
            TimeoutException: If account was not deleted within timeout allowed.
        """

    @abstractmethod
    async def wait_connected(self, timeout_in_seconds=300, interval_in_milliseconds=1000):
        """Waits until API server has connected to the terminal and terminal has connected to the broker.

        Args:
            timeout_in_seconds: Wait timeout in seconds, default is 5m.
            interval_in_milliseconds: Interval between account reloads while waiting for a change, default is 1s.

        Returns:
            A coroutine which resolves when API server is connected to the broker.

        Raises:
            TimeoutException: If account has not connected to the broker within timeout allowed.
        """

    @abstractmethod
    async def connect(self, history_storage: HistoryStorage, history_start_time: datetime = None):
        """Connects to MetaApi.

        Args:
            history_storage: Optional history storage.
            history_start_time: History start time. Used for tests.

        Returns:
            MetaApi connection.
        """

    @abstractmethod
    async def _delay(self, timeout_in_milliseconds):
        pass



from ..metaApi_client import MetaApiClient
from typing_extensions import TypedDict
from typing import Optional


class NewMT4DemoAccount(TypedDict):
    """New MetaTrader 4 demo account model."""
    accountType: Optional[str]
    """Account type."""
    address: Optional[str]
    """Account holder's address."""
    balance: float
    """Account balance."""
    city: Optional[str]
    """Account holder's city."""
    country: Optional[str]
    """Account holder's country."""
    email: str
    """Account holder's email."""
    leverage: float
    """Account leverage."""
    name: Optional[str]
    """Account holder's name."""
    phone: Optional[str]
    """Account holder's phone."""
    serverName: str
    """Server name."""
    state: Optional[str]
    """Account holder's state"""
    zip: Optional[str]
    """Zip address."""


class NewMT5DemoAccount(TypedDict):
    """New MetaTrader 5 demo account model."""
    address: Optional[str]
    """Account holder's address."""
    balance: float
    """Account balance."""
    city: Optional[str]
    """Account holder's city."""
    country: Optional[str]
    """Account holder's country."""
    email: str
    """Account holder's email."""
    languageId: Optional[int]
    """Language id (default is 1)"""
    leverage: float
    """Account leverage."""
    name: Optional[str]
    """Account holder's name."""
    phone: Optional[str]
    """Account holder's phone."""
    serverName: str
    """Server name."""
    state: Optional[str]
    """Account holder's state"""
    zip: Optional[str]
    """Zip address."""


class MetatraderDemoAccountDto(TypedDict):
    """MetaTrader demo account model."""
    login: str
    """Account login."""
    password: str
    """Account password."""
    serverName: str
    """MetaTrader server name."""
    investorPassword: str
    """Account investor (read-only) password."""


class MetatraderDemoAccountClient(MetaApiClient):
    """metaapi.cloud MetaTrader demo account API client."""

    async def create_mt4_demo_account(self, profile_id: str, account: NewMT4DemoAccount) -> MetatraderDemoAccountDto:
        """Creates new MetaTrader 4 demo account.
        Method is accessible only with API access token.

        Args:
            profile_id: Id of the provisioning profile that will be used as the basis for creating this account.
            account: Demo account to create.

        Returns:
            A coroutine resolving with MetaTrader demo account created.
        """
        if self._is_not_jwt_token():
            return self._handle_no_access_exception('create_mt4_demo_account')
        opts = {
            'url': f'{self._host}/users/current/provisioning-profiles/{profile_id}/mt4-demo-accounts',
            'method': 'POST',
            'headers': {
                'auth-token': self._token
            },
            'body': account
        }
        return await self._httpClient.request(opts)

    async def create_mt5_demo_account(self, profile_id: str, account: NewMT5DemoAccount) -> MetatraderDemoAccountDto:
        """Creates new MetaTrader 5 demo account.
        Method is accessible only with API access token.

        Args:
            profile_id: Id of the provisioning profile that will be used as the basis for creating this account.
            account: Demo account to create.

        Returns:
            A coroutine resolving with MetaTrader demo account created.
        """
        if self._is_not_jwt_token():
            return self._handle_no_access_exception('create_mt5_demo_account')
        opts = {
            'url': f'{self._host}/users/current/provisioning-profiles/{profile_id}/mt5-demo-accounts',
            'method': 'POST',
            'headers': {
                'auth-token': self._token
            },
            'body': account
        }
        return await self._httpClient.request(opts)



from .memoryHistoryStorage import MemoryHistoryStorage
from .models import date
from mock import AsyncMock, MagicMock, patch
from datetime import datetime
import pytest
storage = None


@pytest.fixture(autouse=True)
async def run_around_tests(mocker):
    with patch('lib.metaApi.memoryHistoryStorage.HistoryFileManager.start_update_job', new_callable=MagicMock):
        global storage
        storage = MemoryHistoryStorage('accountId')
        await storage.on_connected(1, 1)
        yield


class TestMemoryHistoryStorage:
    @pytest.mark.asyncio
    async def test_load_data_from_file_manager(self):
        """Should load data from the file manager."""
        test_deal = {'id': '37863643', 'type': 'DEAL_TYPE_BALANCE', 'magic': 0, 'time':
                     datetime.fromtimestamp(100).isoformat(),
                     'commission': 0, 'swap': 0, 'profit': 10000, 'platform': 'mt5', 'comment': 'Demo deposit 1'}
        test_order = {'id': '61210463', 'type': 'ORDER_TYPE_SELL', 'state': 'ORDER_STATE_FILLED', 'symbol': 'AUDNZD',
                      'magic': 0, 'time': datetime.fromtimestamp(50).isoformat(), 'doneTime':
                          datetime.fromtimestamp(100).isoformat(), 'currentPrice': 1, 'volume': 0.01,
                      'currentVolume': 0,
                      'positionId': '61206630', 'platform': 'mt5', 'comment': 'AS_AUDNZD_5YyM6KS7Fv:'}
        storage._fileManager.get_history_from_disk = AsyncMock(return_value={'deals': [test_deal],
                                                                             'historyOrders': [test_order]})
        await storage.load_data_from_disk()
        assert storage.deals == [test_deal]
        assert storage.history_orders == [test_order]

    @pytest.mark.asyncio
    async def test_update_disk_storage(self):
        """Should update disk storage."""

        storage._fileManager.update_disk_storage = AsyncMock()
        await storage.update_disk_storage()
        storage._fileManager.update_disk_storage.assert_called()

    @pytest.mark.asyncio
    async def test_return_last_history_order_time(self):
        """Should return last history order time."""

        await storage.on_history_order_added(1, {'id': '1'})
        await storage.on_history_order_added(1, {'id': '2', 'doneTime': date('2020-01-01T00:00:00.000Z')})
        await storage.on_history_order_added(1, {'id': '3', 'doneTime': date('2020-01-02T00:00:00.000Z')})
        assert await storage.last_history_order_time() == date('2020-01-02T00:00:00.000Z')

    @pytest.mark.asyncio
    async def test_return_last_deal_time(self):
        """Should return last deal time."""

        await storage.on_deal_added(1, {'id': '1'})
        await storage.on_deal_added(1, {'id': '2', 'time': date('2020-01-01T00:00:00.000Z')})
        await storage.on_deal_added(1, {'id': '3', 'time': date('2020-01-02T00:00:00.000Z')})
        assert await storage.last_deal_time() == date('2020-01-02T00:00:00.000Z',)

    @pytest.mark.asyncio
    async def test_return_saved_deals(self):
        """Should return saved deals."""

        await storage.on_deal_added(1, {'id': '1', 'time': date('2020-01-01T00:00:00.000Z'), 'type': 'DEAL_TYPE_SELL'})
        await storage.on_deal_added(1, {'id': '7', 'time': date('2020-05-01T00:00:00.000Z'), 'type': 'DEAL_TYPE_BUY'})
        await storage.on_deal_added(1, {'id': '8', 'time': date('2020-02-01T00:00:00.000Z'), 'type': 'DEAL_TYPE_SELL'})
        await storage.on_deal_added(1, {'id': '6', 'time': date('2020-10-01T00:00:00.000Z'), 'type': 'DEAL_TYPE_BUY'})
        await storage.on_deal_added(1, {'id': '4', 'time': date('2020-02-01T00:00:00.000Z'), 'type': 'DEAL_TYPE_SELL'})
        await storage.on_deal_added(1, {'id': '5', 'time': date('2020-06-01T00:00:00.000Z'), 'type': 'DEAL_TYPE_BUY'})
        await storage.on_deal_added(1, {'id': '11', 'type': 'DEAL_TYPE_SELL'})
        await storage.on_deal_added(1, {'id': '3', 'time': date('2020-09-01T00:00:00.000Z'), 'type': 'DEAL_TYPE_BUY'})
        await storage.on_deal_added(1, {'id': '5', 'time': date('2020-06-01T00:00:00.000Z'), 'type': 'DEAL_TYPE_BUY'})
        await storage.on_deal_added(1, {'id': '2', 'time': date('2020-08-01T00:00:00.000Z'), 'type': 'DEAL_TYPE_SELL'})
        await storage.on_deal_added(1, {'id': '10', 'type': 'DEAL_TYPE_SELL'})
        await storage.on_deal_added(1, {'id': '12', 'type': 'DEAL_TYPE_BUY'})

        assert storage.deals == [
            {'id': '10', 'type': 'DEAL_TYPE_SELL'},
            {'id': '11', 'type': 'DEAL_TYPE_SELL'},
            {'id': '12', 'type': 'DEAL_TYPE_BUY'},
            {'id': '1', 'time': date('2020-01-01T00:00:00.000Z'), 'type': 'DEAL_TYPE_SELL'},
            {'id': '4', 'time': date('2020-02-01T00:00:00.000Z'), 'type': 'DEAL_TYPE_SELL'},
            {'id': '8', 'time': date('2020-02-01T00:00:00.000Z'), 'type': 'DEAL_TYPE_SELL'},
            {'id': '7', 'time': date('2020-05-01T00:00:00.000Z'), 'type': 'DEAL_TYPE_BUY'},
            {'id': '5', 'time': date('2020-06-01T00:00:00.000Z'), 'type': 'DEAL_TYPE_BUY'},
            {'id': '2', 'time': date('2020-08-01T00:00:00.000Z'), 'type': 'DEAL_TYPE_SELL'},
            {'id': '3', 'time': date('2020-09-01T00:00:00.000Z'), 'type': 'DEAL_TYPE_BUY'},
            {'id': '6', 'time': date('2020-10-01T00:00:00.000Z'), 'type': 'DEAL_TYPE_BUY'}
        ]

    @pytest.mark.asyncio
    async def test_return_saved_history_orders(self):
        """Should return saved historyOrders."""

        await storage.on_history_order_added(1, {'id': '1', 'doneTime': date('2020-01-01T00:00:00.000Z'),
                                             'type': 'ORDER_TYPE_SELL'})
        await storage.on_history_order_added(1, {'id': '7', 'doneTime': date('2020-05-01T00:00:00.000Z'),
                                             'type': 'ORDER_TYPE_BUY'})
        await storage.on_history_order_added(1, {'id': '8', 'doneTime': date('2020-02-01T00:00:00.000Z'),
                                             'type': 'ORDER_TYPE_SELL'})
        await storage.on_history_order_added(1, {'id': '6', 'doneTime': date('2020-10-01T00:00:00.000Z'),
                                             'type': 'ORDER_TYPE_BUY'})
        await storage.on_history_order_added(1, {'id': '4', 'doneTime': date('2020-02-01T00:00:00.000Z'),
                                             'type': 'ORDER_TYPE_SELL'})
        await storage.on_history_order_added(1, {'id': '5', 'doneTime': date('2020-06-01T00:00:00.000Z'),
                                             'type': 'ORDER_TYPE_BUY'})
        await storage.on_history_order_added(1, {'id': '11', 'type': 'ORDER_TYPE_SELL'})
        await storage.on_history_order_added(1, {'id': '3', 'doneTime': date('2020-09-01T00:00:00.000Z'),
                                             'type': 'ORDER_TYPE_BUY'})
        await storage.on_history_order_added(1, {'id': '5', 'doneTime': date('2020-06-01T00:00:00.000Z'),
                                                 'type': 'ORDER_TYPE_BUY'})
        await storage.on_history_order_added(1, {'id': '2', 'doneTime': date('2020-08-01T00:00:00.000Z'),
                                             'type': 'ORDER_TYPE_SELL'})
        await storage.on_history_order_added(1, {'id': '10', 'type': 'ORDER_TYPE_SELL'})
        await storage.on_history_order_added(1, {'id': '12', 'type': 'ORDER_TYPE_BUY'})

        assert storage.history_orders == [
            {'id': '10', 'type': 'ORDER_TYPE_SELL'},
            {'id': '11', 'type': 'ORDER_TYPE_SELL'},
            {'id': '12', 'type': 'ORDER_TYPE_BUY'},
            {'id': '1', 'doneTime': date('2020-01-01T00:00:00.000Z'), 'type': 'ORDER_TYPE_SELL'},
            {'id': '4', 'doneTime': date('2020-02-01T00:00:00.000Z'), 'type': 'ORDER_TYPE_SELL'},
            {'id': '8', 'doneTime': date('2020-02-01T00:00:00.000Z'), 'type': 'ORDER_TYPE_SELL'},
            {'id': '7', 'doneTime': date('2020-05-01T00:00:00.000Z',), 'type': 'ORDER_TYPE_BUY'},
            {'id': '5', 'doneTime': date('2020-06-01T00:00:00.000Z'), 'type': 'ORDER_TYPE_BUY'},
            {'id': '2', 'doneTime': date('2020-08-01T00:00:00.000Z'), 'type': 'ORDER_TYPE_SELL'},
            {'id': '3', 'doneTime': date('2020-09-01T00:00:00.000Z'), 'type': 'ORDER_TYPE_BUY'},
            {'id': '6', 'doneTime': date('2020-10-01T00:00:00.000Z'), 'type': 'ORDER_TYPE_BUY'}
                                          ]

    @pytest.mark.asyncio
    async def test_return_saved_order_sync_status(self):
        """Should return saved order synchronization status."""

        assert not storage.order_synchronization_finished
        await storage.on_order_synchronization_finished(1, 'synchronizationId')
        assert storage.order_synchronization_finished

    @pytest.mark.asyncio
    async def test_return_saved_deal_sync_status(self):
        """Should return saved deal synchronization status."""

        storage._fileManager.update_disk_storage = AsyncMock()
        assert not storage.deal_synchronization_finished
        await storage.on_deal_synchronization_finished(1, 'synchronizationId')
        storage._fileManager.update_disk_storage.assert_called()
        assert storage.deal_synchronization_finished

    @pytest.mark.asyncio
    async def test_reset_storage(self):
        """Should reset storage."""
        storage._fileManager.delete_storage_from_disk = AsyncMock()
        await storage.on_deal_added(1, {'id': '1', 'time': date('2020-01-01T00:00:00.000Z'), 'type': 'DEAL_TYPE_SELL'})
        await storage.on_history_order_added(1, {'id': '1', 'doneTime': date('2020-01-01T00:00:00.000Z'),
                                             'type': 'ORDER_TYPE_SELL'})
        await storage.clear()
        assert storage.deals == []
        assert storage.history_orders == []
        storage._fileManager.delete_storage_from_disk.assert_called_once()



from ..clients.metaApi.synchronizationListener import SynchronizationListener, HealthStatus
from .reservoir.reservoir import Reservoir
from .models import MetatraderSymbolPrice, date
from typing_extensions import TypedDict
import asyncio
from datetime import datetime
import functools


class ConnectionHealthStatus(TypedDict):
    """Connection health status."""
    connected: bool
    """Flag indicating successful connection to API server."""
    connectedToBroker: bool
    """Flag indicating successful connection to broker."""
    quoteStreamingHealthy: bool
    """Flag indicating that quotes are being streamed successfully from the broker."""
    synchronized: bool
    """Flag indicating a successful synchronization."""
    healthy: bool
    """Flag indicating overall connection health status."""
    message: str
    """Health status message."""


class ConnectionHealthMonitor(SynchronizationListener):
    """Tracks connection health status."""

    def __init__(self, connection):
        """Inits the monitor.

        Args:
            connection: MetaApi connection instance.
        """
        super().__init__()
        self._connection = connection
        self._priceUpdatedAt = None

        async def update_quote_health_job():
            while True:
                await asyncio.sleep(1)
                self._update_quote_health_status()

        async def measure_uptime_job():
            while True:
                await asyncio.sleep(1)
                self._measure_uptime()

        self._quotesHealthy = False
        self._offset = 0
        self._minQuoteInterval = 60
        self._updateQuoteHealthStatusInterval = asyncio.create_task(update_quote_health_job())
        self._measureUptimeInterval = asyncio.create_task(measure_uptime_job())
        self._serverHealthStatus = {}
        self._uptimeReservoirs = {
            '5m': Reservoir(300, 5 * 60 * 1000),
            '1h': Reservoir(600, 60 * 60 * 1000),
            '1d': Reservoir(24 * 60, 24 * 60 * 60 * 1000),
            '1w': Reservoir(24 * 7, 7 * 24 * 60 * 60 * 1000),
        }

    def stop(self):
        """Stops health monitor."""
        self._updateQuoteHealthStatusInterval.cancel()
        self._measureUptimeInterval.cancel()

    async def on_symbol_price_updated(self, instance_index: int, price: MetatraderSymbolPrice):
        """Invoked when a symbol price was updated.

        Args:
            instance_index: Index of an account instance connected.
            price: Updated MetaTrader symbol price.
        """
        try:
            broker_timestamp = date(price['brokerTime']).timestamp()
            self._priceUpdatedAt = datetime.now()
            self._offset = self._priceUpdatedAt.timestamp() - broker_timestamp

        except Exception as err:
            print(f'[{datetime.now().isoformat()}] failed to update quote streaming health status on price ' +
                  'update for account ' + self._connection.account.id, err)

    async def on_health_status(self, instance_index: int, status: HealthStatus):
        """Invoked when a server-side application health status is received from MetaApi.

        Args:
            instance_index: Index of an account instance connected.
            status: Server-side application health status.

        Returns:
            A coroutine which resolves when the asynchronous event is processed.
        """
        self._serverHealthStatus[str(instance_index)] = status

    async def on_disconnected(self, instance_index: int):
        """Invoked when connection to MetaTrader terminal terminated.

        Args:
            instance_index: Index of an account instance connected.

        Returns:
             A coroutine which resolves when the asynchronous event is processed.
        """
        if str(instance_index) in self._serverHealthStatus:
            del self._serverHealthStatus[str(instance_index)]

    @property
    def server_health_status(self) -> HealthStatus:
        """Returns server-side application health status.

        Returns:
            Server-side application health status.
        """
        result = None
        for s in self._serverHealthStatus:
            if not result:
                result = s
            else:
                for field in list(s.keys()):
                    result[field] = result[field] if field in result else s[field]
        return result or {}

    @property
    def health_status(self) -> ConnectionHealthStatus:
        """Returns health status.

        Returns:
            Connection health status.
        """
        status = {
            'connected': self._connection.terminal_state.connected,
            'connectedToBroker': self._connection.terminal_state.connected_to_broker,
            'quoteStreamingHealthy': self._quotesHealthy,
            'synchronized': self._connection.synchronized
        }
        status['healthy'] = status['connected'] and status['connectedToBroker'] and \
            status['quoteStreamingHealthy'] and status['synchronized']
        if status['healthy']:
            message = 'Connection to broker is stable. No health issues detected.'
        else:
            message = 'Connection is not healthy because '
            reasons = []
            if not status['connected']:
                reasons.append('connection to API server is not established or lost')
            if not status['connectedToBroker']:
                reasons.append('connection to broker is not established or lost')
            if not status['synchronized']:
                reasons.append('local terminal state is not synchronized to broker')
            if not status['quoteStreamingHealthy']:
                reasons.append('quotes are not streamed from the broker properly')
            message = message + functools.reduce(lambda a, b: a + ' and ' + b, reasons) + '.'
        status['message'] = message
        return status

    @property
    def uptime(self) -> dict:
        """Returns uptime in percents measured over specific periods of time.

        Returns:
            Uptime in percents measured over specific periods of time.
        """
        uptime = {}
        for key in self._uptimeReservoirs:
            e: Reservoir = self._uptimeReservoirs[key]
            uptime[key] = e.get_statistics()['average']
        return uptime

    def _measure_uptime(self):
        try:
            for key in self._uptimeReservoirs:
                r = self._uptimeReservoirs[key]
                r.push_measurement(100 if (
                    self._connection.terminal_state.connected and self._connection.terminal_state.connected_to_broker
                    and self._connection.synchronized and self._quotesHealthy) else 0)
        except Exception as err:
            print(f'[{datetime.now().isoformat()}] failed to measure uptime for account ' +
                  self._connection.account.id, err)

    def _update_quote_health_status(self):
        try:
            server_date_time = datetime.fromtimestamp(datetime.now().timestamp() - self._offset)
            server_time = server_date_time.strftime('%H:%M:%S.%f')
            day_of_week = server_date_time.isoweekday()
            days_of_week = {
                '1': 'MONDAY',
                '2': 'TUESDAY',
                '3': 'WEDNESDAY',
                '4': 'THURSDAY',
                '5': 'FRIDAY',
                '6': 'SATURDAY',
                '7': 'SUNDAY'
            }
            in_quote_session = False
            if not self._priceUpdatedAt:
                self._priceUpdatedAt = datetime.now()
            if (not self._connection.subscribed_symbols) or (not len(self._connection.subscribed_symbols)):
                self._priceUpdatedAt = datetime.now()
            for symbol in self._connection.subscribed_symbols:
                specification = self._connection.terminal_state.specification(symbol) or {}
                quote_sessions_list = (specification['quoteSessions'] if 'quoteSessions' in specification else [])
                quote_sessions = quote_sessions_list[days_of_week[str(day_of_week)]] if \
                    days_of_week[str(day_of_week)] in quote_sessions_list else []
                for session in quote_sessions:
                    if session['from'] <= server_time <= session['to']:
                        in_quote_session = True
            self._quotesHealthy = (not len(self._connection.subscribed_symbols)) or (not in_quote_session) or \
                                  ((datetime.now().timestamp() - self._priceUpdatedAt.timestamp()) <
                                   self._minQuoteInterval)
        except Exception as err:
            print(f'[{datetime.now().isoformat()}] failed to update quote streaming health status for account ' +
                  self._connection.account.id, err)



from typing_extensions import TypedDict
from typing import List, Optional
from datetime import datetime
from enum import Enum


CopyFactoryStrategySymbolMapping = TypedDict(
    "CopyFactoryStrategySymbolMapping",
    {
        "from": str,  # Symbol name to convert from.
        "to": str  # Symbol name to convert to.
    }
)
"""CopyFactory strategy symbol mapping."""


class CopyFactoryStrategyIdAndName(TypedDict):
    """CopyFactory strategy id and name."""
    id: str
    """Unique strategy id."""
    name: str
    """Human-readable strategy name."""


class CopyFactoryStrategyStopout(TypedDict):
    """CopyFactory strategy stopout."""
    strategy: CopyFactoryStrategyIdAndName
    """Strategy which was stopped out."""
    reason: str
    """Stopout reason. One of yearly-balance, monthly-balance, daily-balance, yearly-equity, monthly-equity,
    daily-equity, max-drawdown"""
    reasonDescription: str
    """Human-readable description of the stopout reason."""
    stoppedAt: datetime
    """Time the strategy was stopped at."""
    stoppedTill: datetime
    """Time the strategy is stopped till."""


class StrategyId(TypedDict):
    """Strategy id"""
    id: str
    """Strategy id"""


class CopyFactoryStrategyStopOutSettings(TypedDict):
    """CopyFactory strategy stopout settings."""
    value: float
    """Value of the stop out risk, expressed as a fraction of 1."""
    startTime: Optional[datetime]
    """The time to start risk calculation from. All previous trades will be ignored. You can use it to reset the risk
    counter after a stopout event."""


class CopyFactoryStrategySymbolFilter(TypedDict):
    """CopyFactory symbol filter."""
    included: List[str]
    """List of symbols copied. Leave the value empty to copy all symbols."""
    excluded: List[str]
    """List of symbols excluded from copying. Leave the value empty to copy all symbols."""


class CopyFactoryStrategyBreakingNewsFilter(TypedDict):
    """CopyFactory breaking news risk filter."""
    priorities: List[str]
    """List of breaking news priorities to stop trading on, leave empty to disable breaking news filter. One of high,
    medium, low."""
    closePositionTimeGapInMinutes: Optional[float]
    """Optional time interval specifying when to force close an already open position before calendar news. Default
    value is 60 minutes."""
    openPositionFollowingTimeGapInMinutes: Optional[float]
    """Optional time interval specifying when it is allowed to open position after calendar news. Default value is
    60 minutes."""


class CopyFactoryStrategyCalendarNewsFilter(TypedDict):
    """CopyFactory calendar news filter."""
    priorities: List[str]
    """List of calendar news priorities to stop trading on, leave empty to disable calendar news filter. One of
    election, high, medium, low."""
    closePositionTimeGapInMinutes: Optional[float]
    """Optional time interval specifying when to force close an already open position before calendar news. Default
    value is 60 minutes."""
    openPositionPrecedingTimeGapInMinutes: Optional[float]
    """Optional time interval specifying when it is still allowed to open position before calendar news. Default value
    is 120 minutes."""
    openPositionFollowingTimeGapInMinutes: Optional[float]
    """Optional time interval specifying when it is allowed to open position after calendar news. Default value is 60
    minutes"""


class CopyFactoryStrategyNewsFilter(TypedDict):
    """CopyFactory news risk filter."""
    breakingNewsFilter: Optional[CopyFactoryStrategyBreakingNewsFilter]
    """Optional breaking news filter."""
    calendarNewsFilter: Optional[CopyFactoryStrategyCalendarNewsFilter]
    """Optional calendar news filter."""


class CopyFactoryStrategyMaxStopLoss(TypedDict):
    """CopyFactory strategy max stop loss settings."""
    value: float
    """Maximum SL value."""
    units: str
    """SL units. Only pips value is supported at this point."""


class CopyFactoryStrategyRiskLimit(TypedDict):
    """CopyFactory risk limit filter."""
    type: str
    """Restriction type. One of daily, monthly, or yearly."""
    applyTo: str
    """Account metric to apply limit to. One of balance, equity."""
    maxRisk: float
    """Max drawdown allowed, expressed as a fraction of 1."""
    closePositions: bool
    """Whether to force close positions when the risk is reached. If value is false then only the new trades will be
    halted, but existing ones will not be closed"""
    startTime: Optional[datetime]
    """Optional time to start risk tracking from. All previous trades will be ignored. You can use this value to reset
    the filter after stopout event."""


class CopyFactoryStrategySubscription(TypedDict):
    """CopyFactory strategy subscriptions."""
    strategyId: str
    """Id of the strategy to subscribe to."""
    multiplier: Optional[float]
    """Optional subscription multiplier, default is 1x."""
    skipPendingOrders: Optional[bool]
    """Optional flag indicating that pending orders should not be copied. Default is to copy pending orders."""
    closeOnly: Optional[str]
    """Optional setting which instructs the application not to open new positions. by-symbol means that it is still
    allowed to open new positions with a symbol equal to the symbol of an existing strategy position (can be used to
    gracefully exit strategies trading in netting mode or placing a series of related trades per symbol). One of
    by-position, by-symbol or leave empty to disable this setting."""
    maxTradeRisk: Optional[float]
    """Optional max risk per trade, expressed as a fraction of 1. If trade has a SL, the trade size will be adjusted to
    match the risk limit. If not, the trade SL will be applied according to the risk limit."""
    reverse: bool
    """Flag indicating that the strategy should be copied in a reverse direction."""
    reduceCorrelations: Optional[str]
    """Optional setting indicating whether to enable automatic trade correlation reduction. Possible settings are not
    specified (disable correlation risk restrictions), by-strategy (limit correlations on strategy level) or by-symbol
    (limit correlations on symbol level)."""
    stopOutRisk: Optional[CopyFactoryStrategyStopOutSettings]
    """Optional stop out setting. All trading will be terminated and positions closed once equity drawdown reaches
    this value."""
    symbolFilter: Optional[CopyFactoryStrategySymbolFilter]
    """Optional symbol filter which can be used to copy only specific symbols or exclude some symbols from copying."""
    newsFilter: Optional[CopyFactoryStrategyNewsFilter]
    """Optional news risk filter configuration."""
    riskLimits: Optional[List[CopyFactoryStrategyRiskLimit]]
    """Optional strategy risk limits. You can configure trading to be stopped once total drawdown generated during
    specific period is exceeded. Can be specified either for balance or equity drawdown."""
    maxStopLoss: Optional[CopyFactoryStrategyMaxStopLoss]
    """Optional stop loss value restriction."""
    maxLeverage: Optional[float]
    """Optional setting indicating maximum leverage allowed when opening a new positions. Any trade which results in a
    higher leverage will be discarded."""
    symbolMapping: Optional[List[CopyFactoryStrategySymbolMapping]]
    """Defines how symbol name should be changed when trading (e.g. when broker uses symbol names with unusual
    suffixes). By default this setting is disabled and the trades are copied using signal source symbol name."""
    tradeSizeScalingMode: Optional[str]
    """If set to balance, the trade size on strategy subscriber will be scaled according to balance to preserve risk.
    If value is none, then trade size will be preserved irregardless of the subscriber balance. If value is
    contractSize, then trade size will be scaled according to contract size. Default is balance."""


class CopyFactoryAccountUpdate(TypedDict):
    """CopyFactory account update."""
    name: str
    """Account human-readable name."""
    connectionId: str
    """Id of the MetaApi MetaTrader account this copy trading account is connected to."""
    reservedMarginFraction: Optional[float]
    """Optional fraction of reserved margin to reduce a risk of margin call. Default is to reserve no margin. We
    recommend using maxLeverage setting instead. Specified as a fraction of balance thus the value is usually greater
    than 1."""
    phoneNumbers: Optional[List[str]]
    """Optional phone numbers to send sms notifications to. Leave empty to receive no sms notifications."""
    minTradeAmount: Optional[float]
    """Optional value of minimal trade size allowed, expressed in amount of account currency. Can be useful if your
    broker charges a fixed fee per transaction so that you can skip small trades with high broker commission rates.
    Default is 100."""
    closeOnly: Optional[str]
    """Optional setting wich instructs the application not to open new positions. by-symbol means that it is still
    allowed to open new positions with a symbol equal to the symbol of an existing strategy position (can be used to
    gracefully exit strategies trading in netting mode or placing a series of related trades per symbol). One of
    by-position, by-symbol or leave empty to disable this setting."""
    stopOutRisk: Optional[CopyFactoryStrategyStopout]
    """Optional stop out setting. All trading will be terminated and positions closed once equity drawdown reaches
    this value."""
    riskLimits: Optional[List[CopyFactoryStrategyRiskLimit]]
    """Optional account risk limits. You can configure trading to be stopped once total drawdown generated during
    specific period is exceeded. Can be specified either for balance or equity drawdown."""
    maxLeverage: Optional[float]
    """Optional setting indicating maxumum leverage allowed when opening a new positions. Any trade which results in a
    higher leverage will be discarded."""
    subscriptions: List[CopyFactoryStrategySubscription]
    """Strategy subscriptions."""


class CopyFactoryAccount(CopyFactoryAccountUpdate):
    """CopyFactory account model."""
    _id: str
    """Account unique identifier."""


class CopyFactoryStrategy(TypedDict):
    """CopyFactory provider strategy"""
    _id: str
    """Unique strategy id."""
    platformCommissionRate: float
    """Commission rate the platform charges for strategy copying, applied to commissions charged by provider. This
    commission applies only to accounts not managed directly by provider. Should be a fraction of 1."""


class CopyFactoryStrategyCommissionScheme(TypedDict):
    """CopyFactory strategy commission scheme."""
    type: str
    """Commission type. One of flat-fee, lots-traded, lots-won, amount-traded, amount-won, high-water-mark."""
    billingPeriod: str
    """Billing period. One of week, month, quarter."""
    commissionRate: float
    """Commission rate. Should be greater than or equal to zero if commission type is flat-fee, lots-traded or
    lots-won, should be greater than or equal to zero and less than or equal to 1 if commission type is amount-traded,
    amount-won, high-water-mark."""


class CopyFactoryStrategyMagicFilter(TypedDict):
    """CopyFactory strategy magic filter."""
    included: List[str]
    """List of magics (expert ids) or magic ranges copied. Leave the value empty to copy all magics."""
    excluded: List[str]
    """List of magics (expert ids) or magic ranges excluded from copying. Leave the value empty to copy all magics."""


class CopyFactoryStrategyTimeSettings(TypedDict):
    """CopyFactory strategy time settings."""
    lifetimeInHours: Optional[float]
    """Optional position lifetime. Default is to keep positions open up to 90 days."""
    openingIntervalInMinutes: Optional[float]
    """Optional time interval to copy new positions. Default is to let 1 minute for the position to get copied. If
    position were not copied during this time, the copying will not be retried anymore."""


class CopyFactoryStrategyUpdate(TypedDict):
    """CopyFactory strategy update."""
    name: str
    """Strategy human-readable name."""
    description: str
    """Longer strategy human-readable description."""
    positionLifecycle: str
    """Position detection mode. Allowed values are netting (single position per strategy per symbol),
    hedging (multiple positions per strategy per symbol)"""
    connectionId: str
    """Id of the MetaApi account providing the strategy."""
    skipPendingOrders: Optional[bool]
    """Optional flag indicating that pending orders should not be copied. Default is to copy pending orders"""
    commissionScheme: Optional[CopyFactoryStrategyCommissionScheme]
    """Commission scheme allowed by this strategy."""
    platformCommissionRate: float
    """Commission rate the platform charges for strategy copying, applied to commissions charged by provider. This
    commission applies only to accounts not managed directly by provider. Should be fraction of 1."""
    maxTradeRisk: Optional[float]
    """Optional max risk per trade, expressed as a fraction of 1. If trade has a SL, the trade size will be adjusted to
    match the risk limit. If not, the trade SL will be applied according to the risk limit."""
    reverse: bool
    """Flag indicating that the strategy should be copied in a reverse direction."""
    reduceCorrelations: Optional[str]
    """Optional setting indicating whether to enable automatic trade correlation reduction. Possible settings are not
    specified (disable correlation risk restrictions), by-strategy (limit correlations on strategy level) or by-symbol
    (limit correlations on symbol level)."""
    stopOutRisk: Optional[CopyFactoryStrategyStopOutSettings]
    """Optional stop out setting. All trading will be terminated and positions closed once equity drawdown reaches
    this value."""
    symbolFilter: Optional[CopyFactoryStrategySymbolFilter]
    """Symbol filters which can be used to copy only specific symbols or exclude some symbols from copying."""
    newsFilter: Optional[CopyFactoryStrategyNewsFilter]
    """News risk filter configuration."""
    riskLimits: Optional[List[CopyFactoryStrategyRiskLimit]]
    """Optional strategy risk limits. You can configure trading to be stopped once total drawdown generated during
    specific period is exceeded. Can be specified either for balance or equity drawdown."""
    maxStopLoss: Optional[CopyFactoryStrategyMaxStopLoss]
    """Optional stop loss value restriction."""
    maxLeverage: Optional[float]
    """Optional max leverage risk restriction. All trades resulting in a leverage value higher than specified will
    be skipped."""
    magicFilter: Optional[CopyFactoryStrategyMagicFilter]
    """Optional magic (expert id) filter."""
    timeSettings: Optional[CopyFactoryStrategyTimeSettings]
    """Settings to manage copying timeframe and position lifetime. Default is to copy position within 1 minute from
    being opened at source and let the position to live for up to 90 days."""
    symbolMapping: Optional[List[CopyFactoryStrategySymbolMapping]]
    """Defines how symbol name should be changed when trading (e.g. when broker uses symbol names with unusual
    suffixes). By default this setting is disabled and the trades are copied using signal source symbol name."""
    tradeSizeScalingMode: Optional[str]
    """If set to balance, the trade size on strategy subscriber will be scaled according to balance to preserve risk.
    If value is none, then trade size will be preserved irregardless of the subscriber balance. If value is
    contractSize, then trade size will be scaled according to contract size. Default is balance."""


class CopyFactorySubscriberOrProvider(TypedDict):
    """CopyFactory provider or subscriber"""
    id: str
    """Profile id."""
    name: str
    """User name."""
    strategies: List[CopyFactoryStrategyIdAndName]
    """Array of strategy IDs provided by provider or subscribed to by subscriber."""


class CopyFactoryTransactionMetrics(TypedDict):
    """Trade copying metrics such as slippage and latencies."""
    tradeCopyingLatency: Optional[float]
    """Trade copying latency, measured in milliseconds based on transaction time provided by broker."""
    tradeCopyingSlippageInBasisPoints: Optional[float]
    """Trade copying slippage, measured in basis points (0.01 percent) based on transaction price provided by broker."""
    tradeCopyingSlippageInAccountCurrency: Optional[float]
    """Trade copying slippage, measured in account currency based on transaction price provided by broker."""
    mtAndBrokerSignalLatency: Optional[float]
    """Trade signal latency introduced by broker and MT platform, measured in milliseconds."""
    tradeAlgorithmLatency: Optional[float]
    """Trade algorithm latency introduced by CopyFactory servers, measured in milliseconds."""
    mtAndBrokerTradeLatency: Optional[float]
    """Trade latency for a copied trade introduced by broker and MT platform, measured in milliseconds"""
    totalLatency: Optional[float]
    """Total trade copying latency, measured in milliseconds. This value might be slightly different from
    tradeCopyingLatency value due to limited measurement precision as it is measured based on timestamps captured
    during copy trading process as opposed to broker data"""


class CopyFactoryTransaction(TypedDict):
    """CopyFactory transaction."""
    id: str
    """Transaction id."""
    type: str
    """Transaction type (one of DEAL_TYPE_BUY, DEAL_TYPE_SELL, DEAL_TYPE_BALANCE, DEAL_TYPE_CREDIT, DEAL_TYPE_CHARGE,
    DEAL_TYPE_CORRECTION, DEAL_TYPE_BONUS, DEAL_TYPE_COMMISSION, DEAL_TYPE_COMMISSION_DAILY,
    DEAL_TYPE_COMMISSION_MONTHLY, DEAL_TYPE_COMMISSION_AGENT_DAILY, DEAL_TYPE_COMMISSION_AGENT_MONTHLY,
    DEAL_TYPE_INTEREST, DEAL_TYPE_BUY_CANCELED, DEAL_TYPE_SELL_CANCELED, DEAL_DIVIDEND, DEAL_DIVIDEND_FRANKED,
    DEAL_TAX). See https://www.mql5.com/en/docs/constants/tradingconstants/dealproperties#enum_deal_type."""
    time: datetime
    """Transaction time."""
    accountId: str
    """CopyFactory account id."""
    symbol: Optional[str]
    """Optional symbol traded."""
    subscriber: CopyFactorySubscriberOrProvider
    """Strategy subscriber."""
    demo: bool
    """Demo account flag."""
    provider: CopyFactorySubscriberOrProvider
    """Strategy provider."""
    strategy: CopyFactoryStrategyIdAndName
    """Strategy."""
    positionId: Optional[str]
    """Source position id."""
    improvement: float
    """High-water mark strategy balance improvement."""
    providerCommission: float
    """Provider commission."""
    platformCommission: float
    """Platform commission."""
    incomingProviderCommission: Optional[float]
    """Commission paid by provider to underlying providers."""
    incomingPlatformCommission: Optional[float]
    """Platform commission paid by provider to underlying providers."""
    lotPrice: Optional[float]
    """Trade lot price."""
    tickPrice: Optional[float]
    """Trade tick price."""
    amount: Optional[float]
    """Trade amount."""
    commission: Optional[float]
    """Trade commission."""
    swap: float
    """Trade swap."""
    profit: float
    """Trade profit."""
    metrics: Optional[CopyFactoryTransactionMetrics]
    """trade copying metrics such as slippage and latencies. Measured selectively for copied trades"""


class ResynchronizationTask(TypedDict):
    """Resynchronization task."""
    _id: str
    """Task unique id."""
    type: str
    """Task type. One of CREATE_ACCOUNT, CREATE_STRATEGY, UPDATE_STRATEGY, REMOVE_STRATEGY."""
    createdAt: datetime
    """The time task was created at."""
    status: str
    """Task status. One of PLANNED, EXECUTING, SYNCHRONIZING."""


class CopyFactoryPortfolioMember(TypedDict):
    """Portfolio strategy member."""
    strategyId: str
    """Member strategy id."""
    multiplier: float
    """Copying multiplier (weight in the portfolio)."""
    skipPendingOrders: Optional[bool]
    """Optional flag indicating that pending orders should not be copied. Default is to copy pending orders."""
    maxTradeRisk: Optional[float]
    """Optional max risk per trade, expressed as a fraction of 1. If trade has a SL, the trade size will be adjusted
    to match the risk limit. If not, the trade SL will be applied according to the risk limit."""
    reverse: bool
    """Flag indicating that the strategy should be copied in a reverse direction."""
    reduceCorrelations: Optional[str]
    """Optional setting indicating whether to enable automatic trade correlation reduction. Possible settings are
    not specified (disable correlation risk restrictions), by-strategy (limit correlations on strategy level) or
    by-symbol (limit correlations on symbol level)."""
    stopOutRisk: Optional[CopyFactoryStrategyStopOutSettings]
    """Optional stop out setting. All trading will be terminated and positions closed once equity drawdown reaches this
    value."""
    symbolFilter: Optional[CopyFactoryStrategySymbolFilter]
    """Symbol filters which can be used to copy only specific symbols or exclude some symbols from copying."""
    newsFilter: Optional[CopyFactoryStrategyNewsFilter]
    """News risk filter configuration."""
    riskLimits: Optional[List[CopyFactoryStrategyRiskLimit]]
    """Optional strategy risk limits. You can configure trading to be stopped once total drawdown generated during
    specific period is exceeded. Can be specified either for balance or equity drawdown."""
    maxStopLoss: Optional[CopyFactoryStrategyMaxStopLoss]
    """Optional stop loss value restriction."""
    maxLeverage: Optional[float]
    """Optional max leverage risk restriction. All trades resulting in a leverage value higher than specified will be
    skipped."""
    symbolMapping: Optional[List[CopyFactoryStrategySymbolMapping]]
    """Defines how symbol name should be changed when trading (e.g. when broker uses symbol names with unusual
    suffixes). By default this setting is disabled and the trades are copied using signal source symbol name."""
    tradeSizeScalingMode: Optional[str]
    """If set to balance, the trade size on strategy subscriber will be scaled according to balance to preserve risk.
    If value is none, then trade size will be preserved irregardless of the subscriber balance. If value is
    contractSize, then trade size will be scaled according to contract size. Default is balance."""


class CopyFactoryPortfolioStrategyUpdate(TypedDict):
    """Portfolio strategy update."""
    name: str
    """Strategy human-readable name."""
    description: str
    """Longer strategy human-readable description."""
    members: List[CopyFactoryPortfolioMember]
    """Array of portfolio memebers."""
    commissionScheme: Optional[CopyFactoryStrategyCommissionScheme]
    """Commission scheme allowed by this strategy. By default monthly billing period with no commission is being
    used."""
    skipPendingOrders: Optional[bool]
    """Optional flag indicating that pending orders should not be copied. Default is to copy pending orders."""
    maxTradeRisk: Optional[float]
    """Optional max risk per trade, expressed as a fraction of 1. If trade has a SL, the trade size will be adjusted
    to match the risk limit. If not, the trade SL will be applied according to the risk limit."""
    reverse: bool
    """Flag indicating that the strategy should be copied in a reverse direction."""
    reduceCorrelations: Optional[str]
    """Optional setting indicating whether to enable automatic trade correlation reduction. Possible settings are not
    specified (disable correlation risk restrictions), by-strategy (limit correlations on strategy level) or by-symbol
    (limit correlations on symbol level)."""
    stopOutRisk: Optional[CopyFactoryStrategyStopOutSettings]
    """Optional stop out setting. All trading will be terminated and positions closed once equity drawdown reaches this
    value."""
    symbolFilter: Optional[CopyFactoryStrategySymbolFilter]
    """Symbol filters which can be used to copy only specific symbols or exclude some symbols from copying."""
    newsFilter: Optional[CopyFactoryStrategyNewsFilter]
    """News risk filter configuration."""
    riskLimits: Optional[List[CopyFactoryStrategyRiskLimit]]
    """Optional strategy risk limits. You can configure trading to be stopped once total drawdown generated during
    specific period is exceeded. Can be specified either for balance or equity drawdown."""
    maxStopLoss: Optional[CopyFactoryStrategyMaxStopLoss]
    """Optional stop loss value restriction."""
    maxLeverage: Optional[float]
    """Optional max leverage risk restriction. All trades resulting in a leverage value higher than specified will be
    skipped."""
    symbolMapping: Optional[List[CopyFactoryStrategySymbolMapping]]
    """Defines how symbol name should be changed when trading (e.g. when broker uses symbol names with unusual
    suffixes). By default this setting is disabled and the trades are copied using signal source symbol name."""
    tradeSizeScalingMode: Optional[str]
    """If set to balance, the trade size on strategy subscriber will be scaled according to balance to preserve risk.
    If value is none, then trade size will be preserved irregardless of the subscriber balance. If value is
    contractSize, then trade size will be scaled according to contract size. Default is balance."""


class CopyFactoryPortfolioStrategy(CopyFactoryPortfolioStrategyUpdate):
    """Portfolio strategy, i.e. the strategy which includes a set of other strategies."""
    _id: str
    """Unique strategy id."""
    platformCommissionRate: float
    """Commission rate the platform charges for strategy copying, applied to commissions charged by provider. This
    commission applies only to accounts not managed directly by provider. Should be fraction of 1."""


class LogLevel(Enum):
    """Log level."""
    INFO = 'INFO'
    WARN = 'WARN'
    ERROR = 'ERROR'


class CopyFactoryUserLogRecord(TypedDict):
    """Trade copying user log record."""
    time: datetime
    """Log record time."""
    level: LogLevel
    """Log level. One of INFO, WARN, ERROR."""
    message: str
    """Log message."""
    symbol: Optional[str]
    """Symbol traded."""
    strategyId: Optional[str]
    """Id of the strategy event relates to."""
    strategyName: Optional[str]
    """Name of the strategy event relates to."""
    positionId: Optional[str]
    """Position id event relates to."""
    side: Optional[str]
    """Side of the trade event relates to. One of buy, sell, close."""
    type: Optional[str]
    """Type of the trade event relates to. One of market, limit, stop."""
    openPrice: Optional[float]
    """Open price for limit and stop orders."""



from asyncio import sleep, gather
from ..metaApi.historyFileManager import HistoryFileManager
from .memoryHistoryStorageModel import MemoryHistoryStorageModel
import pytest
import json
import os
from mock import AsyncMock, patch
from datetime import datetime
from copy import deepcopy
from .models import date
import shutil
file_manager: HistoryFileManager or None = None
storage = None
test_deal = None
test_deal2 = None
test_deal3 = None
test_order = None
test_order2 = None
test_order3 = None
test_config = None


class MockStorage(MemoryHistoryStorageModel):

    def __init__(self):
        super().__init__()
        self._deals = []
        self._historyOrders = []
        self._lastDealTimeByInstanceIndex = {}
        self._lastHistoryOrderTimeByInstanceIndex = {}

    @property
    def deals(self):
        return self._deals

    @property
    def history_orders(self):
        return self._historyOrders

    @property
    def last_deal_time_by_instance_index(self) -> dict:
        return self._lastDealTimeByInstanceIndex

    @property
    def last_history_order_time_by_instance_index(self) -> dict:
        return self._lastHistoryOrderTimeByInstanceIndex

    async def clear(self):
        pass

    def last_deal_time(self):
        pass

    def last_history_order_time(self):
        pass

    def on_deal_added(self, deal):
        pass

    def on_history_order_added(self, order):
        pass


async def read_history_storage_file():
    """Helper function to read saved history storage."""
    history = {
        'deals': [],
        'historyOrders': [],
        'lastDealTimeByInstanceIndex': {},
        'lastHistoryOrderTimeByInstanceIndex': {}
    }
    if os.path.isfile('.metaapi/accountId-application-config.bin'):
        config = json.loads(open('.metaapi/accountId-application-config.bin').read())
        history['lastDealTimeByInstanceIndex'] = config['lastDealTimeByInstanceIndex']
        history['lastHistoryOrderTimeByInstanceIndex'] = config['lastHistoryOrderTimeByInstanceIndex']
    if os.path.isfile('.metaapi/accountId-application-deals.bin'):
        history['deals'] = json.loads(open('.metaapi/accountId-application-deals.bin').read())
    if os.path.isfile('.metaapi/accountId-application-historyOrders.bin'):
        history['historyOrders'] = json.loads(open('.metaapi/accountId-application-historyOrders.bin').read())
    return history


@pytest.fixture(scope="module", autouse=True)
def run_around_module():
    if not os.path.exists('.metaapi'):
        os.mkdir('.metaapi')
    yield
    shutil.rmtree('.metaapi')


@pytest.fixture(autouse=True)
async def run_around_tests():
    global storage
    storage = MockStorage()
    global file_manager
    file_manager = HistoryFileManager('accountId', 'application', storage)
    global test_deal
    test_deal = {'id': '37863643', 'type': 'DEAL_TYPE_BALANCE', 'magic': 0, 'time':
                 datetime.fromtimestamp(100).isoformat(),
                 'commission': 0, 'swap': 0, 'profit': 10000, 'platform': 'mt5', 'comment': 'Demo deposit 1'}
    global test_deal2
    test_deal2 = {'id': '37863644', 'type': 'DEAL_TYPE_SELL', 'magic': 1, 'time':
                  datetime.fromtimestamp(200).isoformat(),
                  'commission': 0, 'swap': 0, 'profit': 10000, 'platform': 'mt5', 'comment': 'Demo deposit 2'}
    global test_deal3
    test_deal3 = {'id': '37863645', 'type': 'DEAL_TYPE_BUY', 'magic': 2, 'time':
                  datetime.fromtimestamp(300).isoformat(),
                  'commission': 0, 'swap': 0, 'profit': 10000, 'platform': 'mt5', 'comment': 'Demo deposit 3'}
    global test_order
    test_order = {'id': '61210463', 'type': 'ORDER_TYPE_SELL', 'state': 'ORDER_STATE_FILLED', 'symbol': 'AUDNZD',
                  'magic': 0, 'time': datetime.fromtimestamp(50).isoformat(), 'doneTime':
                  datetime.fromtimestamp(100).isoformat(), 'currentPrice': 1, 'volume': 0.01, 'currentVolume': 0,
                  'positionId': '61206630', 'platform': 'mt5', 'comment': 'AS_AUDNZD_5YyM6KS7Fv:'}
    global test_order2
    test_order2 = {'id': '61210464', 'type': 'ORDER_TYPE_BUY_LIMIT', 'state': 'ORDER_STATE_FILLED', 'symbol': 'AUDNZD',
                   'magic': 1, 'time': datetime.fromtimestamp(75).isoformat(), 'doneTime':
                   datetime.fromtimestamp(200).isoformat(), 'currentPrice': 1, 'volume': 0.01, 'currentVolume': 0,
                   'positionId': '61206631', 'platform': 'mt5', 'comment': 'AS_AUDNZD_5YyM6KS7Fv:'}
    global test_order3
    test_order3 = {'id': '61210465', 'type': 'ORDER_TYPE_BUY', 'state': 'ORDER_STATE_FILLED', 'symbol': 'AUDNZD',
                   'magic': 2, 'time': datetime.fromtimestamp(100).isoformat(), 'doneTime':
                   datetime.fromtimestamp(300).isoformat(), 'currentPrice': 1, 'volume': 0.01, 'currentVolume': 0,
                   'positionId': '61206632', 'platform': 'mt5', 'comment': 'AS_AUDNZD_5YyM6KS7Fv:'}
    global test_config
    test_config = {
        'lastDealTimeByInstanceIndex': {'0': 1000000000000},
        'lastHistoryOrderTimeByInstanceIndex': {'0': 1000000000010}
    }
    yield
    if os.path.isfile('.metaapi/accountId-application-deals.bin'):
        os.remove('.metaapi/accountId-application-deals.bin')
    if os.path.isfile('.metaapi/accountId-application-historyOrders.bin'):
        os.remove('.metaapi/accountId-application-historyOrders.bin')


class TestHistoryFileManager:

    @pytest.mark.asyncio
    async def test_start_stop_job(self):
        """Should start and stop job."""

        with patch.object(file_manager, 'update_disk_storage', new_callable=AsyncMock):
            with patch('lib.metaApi.historyFileManager.asyncio.sleep', new=lambda x: sleep(x/120)):
                file_manager.update_disk_storage = AsyncMock()
                file_manager.start_update_job()
                await sleep(0.6)
                assert file_manager.update_disk_storage.call_count == 1
                await sleep(0.6)
                assert file_manager.update_disk_storage.call_count == 2
                file_manager.stop_update_job()
                await sleep(0.6)
                assert file_manager.update_disk_storage.call_count == 2
                file_manager.start_update_job()
                await sleep(0.7)
                assert file_manager.update_disk_storage.call_count == 3
                file_manager.stop_update_job()

    @pytest.mark.asyncio
    async def test_read_history(self):
        """Should read history from file."""

        f = open('.metaapi/accountId-application-deals.bin', "w+")
        f.write(json.dumps([test_deal]))
        f.close()

        f = open('.metaapi/accountId-application-historyOrders.bin', "w+")
        f.write(json.dumps([test_order]))
        f.close()

        f = open('.metaapi/accountId-application-config.bin', "w+")
        f.write(json.dumps(test_config))
        f.close()

        history = await file_manager.get_history_from_disk()
        expected_deal = deepcopy(test_deal)
        expected_deal['time'] = date(expected_deal['time'])
        assert history['deals'] == [expected_deal]
        expected_order = deepcopy(test_order)
        expected_order['time'] = date(expected_order['time'])
        expected_order['doneTime'] = date(expected_order['doneTime'])
        assert history['historyOrders'] == [expected_order]
        assert history['lastDealTimeByInstanceIndex'] == test_config['lastDealTimeByInstanceIndex']
        assert history['lastHistoryOrderTimeByInstanceIndex'] == test_config['lastHistoryOrderTimeByInstanceIndex']

    @pytest.mark.asyncio
    async def test_save_items(self):
        """Should save items in a file."""

        storage._deals = [test_deal]
        storage._historyOrders = [test_order]
        storage._lastDealTimeByInstanceIndex = {'0': 1000000000000}
        storage._lastHistoryOrderTimeByInstanceIndex = {'0': 1000000000010}
        file_manager.set_start_new_deal_index(0)
        file_manager.set_start_new_order_index(0)
        await file_manager.update_disk_storage()
        saved_data = await read_history_storage_file()
        assert saved_data['deals'] == [test_deal]
        assert saved_data['historyOrders'] == [test_order]
        assert saved_data['lastDealTimeByInstanceIndex'] == {'0': 1000000000000}
        assert saved_data['lastHistoryOrderTimeByInstanceIndex'] == {'0': 1000000000010}

    @pytest.mark.asyncio
    async def test_replace_nth_item(self):
        """Should replace Nth item in a file."""

        storage._deals = [test_deal, test_deal2]
        storage._historyOrders = [test_order, test_order2]
        file_manager.set_start_new_deal_index(0)
        file_manager.set_start_new_order_index(0)
        await file_manager.update_disk_storage()
        test_deal2['magic'] = 100
        test_order2['magic'] = 100
        file_manager.set_start_new_deal_index(1)
        file_manager.set_start_new_order_index(1)
        await file_manager.update_disk_storage()
        saved_data = await read_history_storage_file()
        assert saved_data['deals'] == [test_deal, test_deal2]
        assert saved_data['historyOrders'] == [test_order, test_order2]

    @pytest.mark.asyncio
    async def test_replace_all_items(self):
        """Should replace all items in a file."""

        storage._deals = [test_deal, test_deal2]
        storage._historyOrders = [test_order, test_order2]
        file_manager.set_start_new_deal_index(0)
        file_manager.set_start_new_order_index(0)
        await file_manager.update_disk_storage()
        test_deal['magic'] = 100
        test_deal2['magic'] = 100
        test_order['magic'] = 100
        test_order2['magic'] = 100
        file_manager.set_start_new_deal_index(0)
        file_manager.set_start_new_order_index(0)
        await file_manager.update_disk_storage()
        saved_data = await read_history_storage_file()
        assert saved_data['deals'] == [test_deal, test_deal2]
        assert saved_data['historyOrders'] == [test_order, test_order2]

    @pytest.mark.asyncio
    async def test_append_new_object(self):
        """Should append a new object to already saved ones."""

        storage._deals = [test_deal, test_deal2]
        storage._historyOrders = [test_order, test_order2]
        file_manager.set_start_new_deal_index(0)
        file_manager.set_start_new_order_index(0)
        await file_manager.update_disk_storage()
        storage._deals = [test_deal, test_deal2, test_deal3]
        storage._historyOrders = [test_order, test_order2, test_order3]
        file_manager.set_start_new_deal_index(2)
        file_manager.set_start_new_order_index(2)
        await file_manager.update_disk_storage()
        saved_data = await read_history_storage_file()
        assert saved_data['deals'] == [test_deal, test_deal2, test_deal3]
        assert saved_data['historyOrders'] == [test_order, test_order2, test_order3]

    @pytest.mark.asyncio
    async def test_not_corrupt(self):
        """Should not corrupt the disk storage if update called multiple times."""
        storage._deals = [test_deal, test_deal2]
        storage._historyOrders = [test_order, test_order2]
        storage._lastDealTimeByInstanceIndex = {'0': 1000000000000}
        storage._lastHistoryOrderTimeByInstanceIndex = {'0': 1000000000010}
        file_manager.set_start_new_deal_index(0)
        file_manager.set_start_new_order_index(0)
        await file_manager.update_disk_storage()
        storage._deals = [test_deal, test_deal2, test_deal3]
        storage._historyOrders = [test_order, test_order2, test_order3]
        storage._lastDealTimeByInstanceIndex = {'0': 1000000000000}
        storage._lastHistoryOrderTimeByInstanceIndex = {'0': 1000000000010}
        file_manager.set_start_new_deal_index(2)
        file_manager.set_start_new_order_index(2)
        await gather(*[
            file_manager.update_disk_storage(),
            file_manager.update_disk_storage(),
            file_manager.update_disk_storage(),
            file_manager.update_disk_storage(),
            file_manager.update_disk_storage()
        ])
        json.loads(open('.metaapi/accountId-application-historyOrders.bin').read())
        json.loads(open('.metaapi/accountId-application-deals.bin').read())
        json.loads(open('.metaapi/accountId-application-config.bin').read())

    @pytest.mark.asyncio
    async def test_remove_history(self):
        """Should remove history from disk."""

        open('.metaapi/accountId-application-deals.bin', "w+").close()
        open('.metaapi/accountId-application-historyOrders.bin', "w+").close()
        open('.metaapi/accountId-application-config.bin', "w+").close()
        assert os.path.isfile('.metaapi/accountId-application-deals.bin')
        assert os.path.isfile('.metaapi/accountId-application-historyOrders.bin')
        assert os.path.isfile('.metaapi/accountId-application-config.bin')
        await file_manager.delete_storage_from_disk()
        assert not os.path.isfile('.metaapi/accountId-application-deals.bin')
        assert not os.path.isfile('.metaapi/accountId-application-historyOrders.bin')
        assert not os.path.isfile('.metaapi/accountId-application-config.bin')

    @pytest.mark.asyncio
    async def test_remove_history_if_not_exists(self):
        """Should do nothing on remove history if files don't exist."""

        assert not os.path.isfile('.metaapi/accountId-application-deals.bin')
        assert not os.path.isfile('.metaapi/accountId-application-historyOrders.bin')
        assert not os.path.isfile('.metaapi/accountId-application-config.bin')
        await file_manager.delete_storage_from_disk()



from .metaApi.metaApi import MetaApi
from .metaApi.historyStorage import HistoryStorage
from .metaApi.memoryHistoryStorage import MemoryHistoryStorage
from .clients.metaApi.synchronizationListener import SynchronizationListener
from .copyFactory.copyFactory import CopyFactory



from ..metaApi_client import MetaApiClient
from .copyFactory_models import CopyFactoryStrategyStopout, CopyFactoryUserLogRecord
from typing import List
from httpx import Response
from datetime import datetime
from ...metaApi.models import date, format_date


class TradingClient(MetaApiClient):
    """metaapi.cloud CopyFactory history API (trade copying history API) client (see
    https://trading-api-v1.project-stock.agiliumlabs.cloud/swagger/#/)"""

    def __init__(self, http_client, token: str, domain: str = 'agiliumtrade.agiliumtrade.ai'):
        """Inits CopyFactory history API client instance.

        Args:
            http_client: HTTP client.
            token: Authorization token.
            domain: Domain to connect to, default is agiliumtrade.agiliumtrade.ai.
        """
        super().__init__(http_client, token, domain)
        self._host = f'https://trading-api-v1.{domain}'

    async def resynchronize(self, account_id: str, strategy_ids: List[str] = None) -> Response:
        """Resynchronizes the account. See
        https://trading-api-v1.agiliumtrade.agiliumtrade.ai/swagger/#!/default
        /post_users_current_accounts_accountId_resynchronize

        Args:
            account_id: Account id.
            strategy_ids: Optional array of strategy ids to recynchronize. Default is to synchronize all strategies.

        Returns:
            A coroutine which resolves when resynchronization is scheduled.
        """
        if self._is_not_jwt_token():
            return self._handle_no_access_exception('resynchronize')
        qs = {}
        if strategy_ids:
            qs['strategyId'] = strategy_ids
        opts = {
          'url': f'{self._host}/users/current/accounts/{account_id}/resynchronize',
          'method': 'POST',
          'headers': {
            'auth-token': self._token
          },
          'params': qs,
        }
        return await self._httpClient.request(opts)

    async def get_stopouts(self, account_id: str) -> 'Response[List[CopyFactoryStrategyStopout]]':
        """Returns subscriber account stopouts. See
        https://trading-api-v1.agiliumtrade.agiliumtrade.ai/swagger/#!/default
        /get_users_current_accounts_accountId_stopouts

        Args:
            account_id: Account id.

        Returns:
            A coroutine which resolves with stopouts found.
        """
        if self._is_not_jwt_token():
            return self._handle_no_access_exception('get_stopouts')
        opts = {
            'url': f'{self._host}/users/current/accounts/{account_id}/stopouts',
            'method': 'GET',
            'headers': {
                'auth-token': self._token
            }
        }
        return await self._httpClient.request(opts)

    async def reset_stopouts(self, account_id: str, strategy_id: str, reason: str) -> Response:
        """Resets strategy stopouts. See
        https://trading-api-v1.agiliumtrade.agiliumtrade.ai/swagger/#!/default/post_users_current_accounts_accountId_
        strategies_subscribed_strategyId_stopouts_reason_reset


        Args:
            account_id: Account id.
            strategy_id: Strategy id.
            reason: Stopout reason to reset. One of yearly-balance, monthly-balance, daily-balance, yearly-equity,
            monthly-equity, daily-equity, max-drawdown.

        Returns:
            A coroutine which resolves when the stopouts are reset.
        """
        if self._is_not_jwt_token():
            return self._handle_no_access_exception('reset_stopouts')
        opts = {
            'url': f'{self._host}/users/current/accounts/{account_id}/strategies-subscribed/{strategy_id}' +
                   f'/stopouts/{reason}/reset',
            'method': 'POST',
            'headers': {
                'auth-token': self._token
            }
        }
        return await self._httpClient.request(opts)

    async def get_user_log(self, account_id: str, start_time: datetime = None, end_time: datetime = None,
                           offset: int = 0, limit: int = 1000) -> 'Response[List[CopyFactoryUserLogRecord]]':
        """Returns copy trading user log for an account and time range. See
        https://trading-api-v1.project-stock.v2.agiliumlabs.cloud/swagger/#!/default
        /get_users_current_accounts_accountId_user_log


        Args:
            account_id: Account id.
            start_time: Time to start loading data from.
            end_time: Time to stop loading data at.
            offset: Pagination offset. Default is 0.
            limit: Pagination limit. Default is 1000.

        Returns:
            A coroutine which resolves with log records found.
        """
        if self._is_not_jwt_token():
            return self._handle_no_access_exception('get_user_log')
        qs = {
            'offset': offset,
            'limit': limit
        }
        if start_time:
            qs['startTime'] = format_date(start_time)
        if end_time:
            qs['endTime'] = format_date(end_time)
        opts = {
            'url': f'{self._host}/users/current/accounts/{account_id}/user-log',
            'method': 'GET',
            'headers': {
                'auth-token': self._token
            },
            'params': qs
        }
        result = await self._httpClient.request(opts)
        if result and isinstance(result, List):
            for r in result:
                r['time'] = date(r['time'])
        return result



from mock import AsyncMock, MagicMock
from ..clients.metaApi.metatraderDemoAccount_client import MetatraderDemoAccountClient, NewMT4DemoAccount, \
    NewMT5DemoAccount, MetatraderDemoAccountDto
from .metatraderDemoAccountApi import MetatraderDemoAccountApi
from .metatraderDemoAccount import MetatraderDemoAccount
import pytest


class MockClient(MetatraderDemoAccountClient):
    def create_mt4_demo_account(self, profile_id: str, account: NewMT4DemoAccount) -> MetatraderDemoAccountDto:
        pass

    def create_mt5_demo_account(self, profile_id: str, account: NewMT5DemoAccount) -> MetatraderDemoAccountDto:
        pass


client = MockClient(MagicMock(), MagicMock())
api = MetatraderDemoAccountApi(client)


@pytest.fixture(autouse=True)
async def run_around_tests():
    global api
    api = MetatraderDemoAccountApi(client)
    yield


class TestMetatraderAccountApi:
    @pytest.mark.asyncio
    async def test_create_mt4(self):
        """Should create MT4 demo account."""
        expected = {
            'login': '12345',
            'password': 'qwerty',
            'serverName': 'HugosWay-Demo3',
            'investorPassword': 'qwerty'
        }
        client.create_mt4_demo_account = AsyncMock(return_value=expected)
        api = MetatraderDemoAccountApi(client)
        new_account_data = {
            'balance': 10,
            'email': 'test@test.com',
            'leverage': 15,
            'serverName': 'server'
        }
        account = await api.create_mt4_demo_account('profileId1', new_account_data)
        assert account.login == expected['login']
        assert account.password == expected['password']
        assert account.server_name == expected['serverName']
        assert account.investor_password == expected['investorPassword']
        assert isinstance(account, MetatraderDemoAccount)
        client.create_mt4_demo_account.assert_called_with('profileId1', new_account_data)

    @pytest.mark.asyncio
    async def test_create_mt5(self):
        """Should create MT5 demo account."""
        expected = {
            'login': '12345',
            'password': 'qwerty',
            'serverName': 'HugosWay-Demo3',
            'investorPassword': 'qwerty'
        }
        client.create_mt5_demo_account = AsyncMock(return_value=expected)
        api = MetatraderDemoAccountApi(client)
        new_account_data = {
            'balance': 10,
            'email': 'test@test.com',
            'leverage': 15,
            'serverName': 'server'
        }
        account = await api.create_mt5_demo_account('profileId2', new_account_data)
        assert account.login == expected['login']
        assert account.password == expected['password']
        assert account.server_name == expected['serverName']
        assert account.investor_password == expected['investorPassword']
        assert isinstance(account, MetatraderDemoAccount)
        client.create_mt5_demo_account.assert_called_with('profileId2', new_account_data)
